import os
import argparse
from readpaf import parse_paf
import pysam

DEFAULT_KMER_SIZE = 9
DEFAULT_SIG_MOVE_OFFSET = 0
DNA_STRIDE = 5
RNA_STRIDE = 10

profile_dic = {
        "guppy_dna_r9.4.1_450bps_fast_prom": [3, 2],
        "guppy_dna_r9.4.1_450bps_hac_prom": [3, 2],
        "guppy_dna_r9.4.1_450bps_sup_prom": [4, 3],
        "guppy_dna_r10.4.1_e8.2_400bps_fast": [1, 0],
        "guppy_dna_r10.4.1_e8.2_400bps_hac": [1, 0],
        "guppy_dna_r10.4.1_e8.2_400bps_sup": [1, 0]}

def list_profiles():
    # print(profile_dic)
    print("{}\t{}\t{}".format("name", "kmer_length", "sig_move_offset"))
    for profile in profile_dic:
        print("{}\t{}\t{}".format(profile, profile_dic[profile][0], profile_dic[profile][1]))
    print("If the profile you wanted is not listed here, please refer calculate_offsets.md on github to learn how to generate these values for your new data.")

def search_for_profile(profile):
    if profile in profile_dic:
        return profile_dic[profile]
    else:
        raise Exception("Error: specified profile ({}) is not found. Please run reform with -k 1 -s 0. Then run calculate_offsets.py and rerun reform with the recommended kmer_length and sig_move_offset.".format(profile))
def run(args):
    if args.list_profile:
        list_profiles()
        return
    else:
        if args.bam == "":
            raise Exception("Error: the following argument is required: -b/--bam")
        if args.output == "":
            raise Exception("Error: the following argument is required: -o/--output")

    if args.profile == "":
        kmer_length = args.kmer_length
        sig_move_offset = args.sig_move_offset
    else:
        kmer_length, sig_move_offset = search_for_profile(args.profile)
    if kmer_length < 1:
        raise Exception("Error: kmer length must be a positive integer")
    if sig_move_offset < 0:
        raise Exception("Error: signal move offset must not be less than zero")

    if kmer_length <= sig_move_offset:
        raise Exception("Error: signal move offset value must be smaller than the kmer length.")

    if (args.c and args.output[-4:] != ".paf") or (not args.c and args.output[-4:] != ".tsv"):
        raise Exception("Error: please provide the output file with correct extension (.tsv/.paf)")

    print("kmer_length: " + str(kmer_length))
    print("sig_move_offset: " + str(sig_move_offset))
    print("input bam: " + args.bam)
    if args.c:
        print("output format: " + "paf")
    else:
        print("output format: " + "tsv")
    print("output file: " + args.output)

    stride = DNA_STRIDE
    if args.rna:
        stride = RNA_STRIDE
    print("Info: Default stride: {}".format(stride))

    samfile = pysam.AlignmentFile(args.bam, mode='r', check_sq=False)
    fout = open(args.output, "w")
    processed_sam_record_count = 0

    for sam_record in samfile:

        len_seq = len(sam_record.get_forward_sequence()) - kmer_length + 1 # to get the number of kmers

        if not sam_record.has_tag("ns"):
            raise Exception("Error: tag '{}' is not found. Please check your input SAM/BAM file.".format("ns"))
        if not sam_record.has_tag("ts"):
            raise Exception("Error: tag '{}' is not found. Please check your input SAM/BAM file.".format("ts"))
        if not sam_record.has_tag("mv"):
            raise Exception("Error: tag '{}' is not found. Please check your input SAM/BAM file.".format("mv"))

        ns = int(sam_record.get_tag("ns"))
        ts = int(sam_record.get_tag("ts"))
        mv = sam_record.get_tag("mv")

        # print("ns: " + str(ns))
        # print("ts: " + str(ts))
        # print(mv[:5])

        len_mv = len(mv)
        if len_mv == 0:
            raise Exception("Error: mv array length is 0.")

        if mv[0] != stride:
            print("Info: Found stride to be {}, this value will be used.".format(mv[0]))
        stride = int(mv[0])
        if not args.c:
            move_count = 0
            i = 1
            while move_count < sig_move_offset + 1:
                value = mv[i]
                if value == 1:
                    move_count += 1
                i += 1
            end_idx = ts + (i - 1) * stride
            start_idx = end_idx - stride
            kmer_idx = 0
            if args.rna:
                kmer_idx = len_seq - 1

            while i < len_mv:
                value = mv[i]
                if len_seq > 0 and value == 1:
                    fout.write("{}\t".format(sam_record.query_name))
                    fout.write("{}\t".format(kmer_idx))
                    fout.write("{}\t".format(start_idx))
                    fout.write("{}\n".format(end_idx))
                    start_idx = end_idx
                    if args.rna:
                        kmer_idx -= 1
                    else:
                        kmer_idx += 1

                    len_seq -= 1
                if len_seq > 0 and i == len_mv-1:
                    fout.write("{}\t".format(sam_record.query_name))
                    fout.write("{}\t".format(kmer_idx))
                    fout.write("{}\t".format(start_idx))
                    fout.write("{}\n".format(ns))
                    start_idx = ns
                    if args.rna:
                        kmer_idx -= 1
                    else:
                        kmer_idx += 1
                    len_seq -= 1

                end_idx = end_idx + stride
                i += 1
            if len_seq != 0:
                raise Exception("Error: error in the implementation. Please report the command with minimal reproducible data. Read_id: {}".format(sam_record.query_name));

        # write paf format
        else:
            fout.write("{}\t".format(sam_record.query_name)) #1 read_id
            fout.write("{}\t".format(ns)) #2 Raw signal length (number of samples)
            move_count = 0
            i = 1
            start_idx = 0
            kmer_idx = 0

            while move_count < sig_move_offset + 1:
                value = mv[i]
                if value == 1:
                    move_count += 1
                    start_idx = i
                i += 1

            fout.write("{}\t".format(ts + (i - 2) * stride)) #3 Raw signal start index (0-based; BED-like; closed)

            j = 1
            l_end_raw = 0
            len_seq_1 = len_seq + sig_move_offset + 1
            end_idx = j + 1
            while j < len_mv:
                value = mv[j]
                if len_seq_1 > 0 and value == 1:
                    len_seq_1 -= 1
                    end_idx = j
                j += 1
            if len_seq_1 > 0 and j == len_mv:
                l_end_raw = ns
            else:
                l_end_raw = ts + (end_idx-1) * stride

            fout.write("{}\t".format(l_end_raw)) #4 Raw signal end index (0-based; BED-like; open)
            fout.write("{}\t".format("+")) #5 Relative strand: “+” or “-“
            fout.write("{}\t".format(sam_record.query_name)) #6 Same as column 1
            fout.write("{}\t".format(len_seq)) #7 base-called sequence length (no. of k-mers)

            if args.rna:
                fout.write("{}\t".format(len_seq))  # 8 k-mer start index on basecalled sequence (0-based)
                fout.write("{}\t".format(kmer_idx))  # 9 k-mer end index on basecalled sequence (0-based)
            else:
                fout.write("{}\t".format(kmer_idx))  # 8 k-mer start index on basecalled sequence (0-based)
                fout.write("{}\t".format(len_seq))  # 9 k-mer end index on basecalled sequence (0-based)
            fout.write("{}\t".format(len_seq - kmer_idx)) #10 Number of k-mers matched on basecalled sequence
            fout.write("{}\t".format(len_seq)) #11 Same as column 7
            fout.write("{}\t".format("255")) #12 Mapping quality (0-255; 255 for missing)
            fout.write("{}".format("ss:Z:")) #12 Mapping quality (0-255; 255 for missing)

            while i < len_mv:
                value = mv[i]
                # print("{}\t{}".format(i, value))
                if len_seq > 0 and value == 1:
                    fout.write("{},".format((i-start_idx) * stride))  # ss
                    start_idx = i
                    len_seq -= 1
                if len_seq > 0 and i == len_mv-1:
                    if (ns - ((i-1) * stride + ts)) < 0:
                        raise Exception("Error: error in calcuation. (ns - ((i-1)*EXPECTED_STRIDE + ts)) > 0 is not valid")
                    len_seq -= 1
                    l_duration = ((i-start_idx) * stride) + (ns - ((i - 1) * stride + ts))
                    fout.write("{},".format(l_duration))  # ss
                i += 1

            if len_seq != 0:
                raise Exception("Error: error in the implementation. Please report the command with minimal reproducible data. Read_id: {}".format(sam_record.query_name));

            fout.write("{}".format("\n"))  # newline
        processed_sam_record_count += 1

    samfile.close()
    fout.close()
    print("processed_sam_record_count: " + str(processed_sam_record_count))


def argparser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument('-k', '--kmer_length', required=False, default=DEFAULT_KMER_SIZE, type=int, help="kmer length")
    parser.add_argument('-m', '--sig_move_offset', required=False, default=DEFAULT_SIG_MOVE_OFFSET, type=int, help="signal move offset")
    parser.add_argument('-c', action='store_true', help="write move table in paf format")
    parser.add_argument('-b', '--bam', required=False, default="", type=str, help="input SAM/BAM file produced by the basecaller")
    parser.add_argument('-o', '--output', required=False, default="", type=str, help="output .tsv/.paf file")
    parser.add_argument('--rna', required=False, action='store_true', help="specify for RNA reads")
    parser.add_argument('--profile', required=False, default="", type=str, help="determine -k and -m values using preset values")
    parser.add_argument('--list_profile', action='store_true', help="list the available profiles")
    return parser

if __name__ == "__main__":
    parser = argparser()
    args = parser.parse_args()
    try:
        run(args)
    except Exception as e:
        print(str(e))
        exit(1)



from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from src import plot, reform, realign, plot_pileup, plot_tracks, calculate_offsets
from ._version import __version__

modules = ['plot', 'reform', 'realign', 'plot_pileup', 'plot_tracks', 'calculate_offsets']
version = "squigualiser     {}".format(__version__)

def main():
    parser = ArgumentParser(
        'squigualiser',
        formatter_class=ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        '-v', '--version', action='version',
        version='%(prog)s {}'.format(__version__)
    )

    subparsers = parser.add_subparsers(
        title='subcommands', description='valid commands',
        help='additional help', dest='command'
    )
    subparsers.required = True
    for module in modules:
        mod = globals()[module]
        p = subparsers.add_parser(module, parents=[mod.argparser()])
        p.set_defaults(func=mod.run)

    args = parser.parse_args()
    args.func(args)



import html
from http import HTTPStatus
import http.server
import io
import socketserver
import os
import sys
from urllib.parse import quote, unquote
import plot
import plot_pileup

class SquigHttpRequestHandler(http.server.SimpleHTTPRequestHandler):
    def do_GET(self):
        if self.path == '/home':
            self.path = 'src/server/home.html'

        return http.server.SimpleHTTPRequestHandler.do_GET(self)

    def do_POST(self):
        if self.path == "/generate_plots":
            content_len = int(self.headers.get('content-length', 0))
            post_data_bytes = self.rfile.read(content_len)
            post_data_parts = post_data_bytes.decode("UTF-8").split()
            use_pileup = post_data_parts[0]
            plot_params = post_data_parts[1:]
            print(use_pileup, plot_params)
            
            try:
                if use_pileup == "true":
                    parser = plot_pileup.argparser()
                    args = parser.parse_args(plot_params)
                    plot_pileup.run(args)
                else:
                    parser = plot.argparser()
                    args = parser.parse_args(plot_params)
                    plot.run(args)
                self.send_response(200)
                self.send_header('Content-type','text/html')
                self.end_headers()
                self.wfile.write(bytes("Success", "utf8"))
            except Exception as e:
                self.send_response(400)
                self.send_header('Content-type','text/html')
                self.end_headers()
                self.wfile.write(bytes(str(e), "utf8"))
            except SystemExit:
                self.send_response(400)
                self.send_header('Content-type','text/html')
                self.end_headers()
                self.wfile.write(bytes("Error: please check the arguments again.", "utf8"))

        return
    
    def list_directory(self, path):
        """ Overridden method to have a custom formatting for the directory listing html page.
            Presents ".." to go back to previous directory.

        """
        try:
            list = os.listdir(path)
        except OSError:
            self.send_error(
                HTTPStatus.NOT_FOUND,
                "No permission to list directory")
            return None
        list.sort(key=lambda a: a.lower())
        r = []
        try:
            displaypath = unquote(self.path,
                                               errors='surrogatepass')
        except UnicodeDecodeError:
            displaypath = unquote(path)
        displaypath = html.escape(displaypath, quote=False)
        enc = sys.getfilesystemencoding()
        title = 'Directory listing for %s' % displaypath
        r.append('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" '
                 '"http://www.w3.org/TR/html4/strict.dtd">')
        r.append('<html>\n<head>')
        r.append('<meta http-equiv="Content-Type" '
                 'content="text/html; charset=%s">' % enc)
        r.append('<link rel="stylesheet" href="/src/server/style.css">')
        r.append('<title>%s</title>\n</head>' % title)
        r.append('<body>')
        r.append('<ul>')

        parent_path = '/'.join(self.path.split('/')[0:-2])
        if parent_path == '':
            parent_path = '/'
        if self.path != '/':
            r.append('<li><a href="%s">..</a></li>'
                        % (quote(parent_path,
                                            errors='surrogatepass')))
            
        for name in list:
            fullname = os.path.join(path, name)
            displayname = linkname = name
            # Append / for directories or @ for symbolic links
            if os.path.isdir(fullname):
                displayname = name + "/"
                linkname = name + "/"
            if os.path.islink(fullname):
                displayname = name + "@"
                # Note: a link to a directory displays with @ and links with /

            if os.path.isfile(fullname) :
                r.append('<li><a target="_blank" href="%s">%s</a></li>'
                        % (quote(linkname,
                                            errors='surrogatepass'),
                        html.escape(displayname, quote=False)))
            else:
                r.append('<li><a href="%s">%s</a></li>'
                    % (quote(linkname,
                                        errors='surrogatepass'),
                    html.escape(displayname, quote=False)))
        r.append('</ul>\n</body>\n</html>\n')
        encoded = '\n'.join(r).encode(enc, 'surrogateescape')
        f = io.BytesIO()
        f.write(encoded)
        f.seek(0)
        self.send_response(HTTPStatus.OK)
        self.send_header("Content-type", "text/html; charset=%s" % enc)
        self.send_header("Content-Length", str(len(encoded)))
        self.end_headers()
        return f

handler_object = SquigHttpRequestHandler

PORT = 8000
socketserver.TCPServer.allow_reuse_address=True
my_server = socketserver.TCPServer(("", PORT), handler_object)
print(f'Link to access Squigualiser on browser: http://localhost:{PORT}/home')

my_server.serve_forever()


"""
Signal to seQuence alignment Plot - plot
Hiruna Samarakoon - Garvan Medical Institute
hiruna@unsw.edu.au
"""
from bokeh.plotting import output_file, save
from bokeh.layouts import column
import argparse
import os
# import plot
from src import plot_pileup

PLOT_HEIGHT_SCALE = 1000
def run(args):
    print("Info: commands file: {}".format(args.file))
    commands_file = open(args.file, "r")
    num_commands = int(commands_file.readline().strip().split('=')[1])
    print("Info: no. of commands: {}".format(num_commands))

    plot_heights = commands_file.readline().strip().split('=')[1].split(',')
    if plot_heights[0] == '*' or args.auto_height:
        args.auto_height = True
        print("Info: track heights will be set automatically.")
    if not args.auto_height:
        if len(plot_heights) != num_commands:
            raise Exception("Error: number of commands ({}) does not match the number of heights specified ({})".format(num_commands, len(plot_heights)))
        print("Info: specified plot heights: {}".format(str(plot_heights)))

    num_tracks = 0
    pileup = []
    num_plots = []
    total_plots = 0

    for i in range(0, num_commands):
        line_ = commands_file.readline().strip().split()
        tool = str(line_[1])
        # print(tool)
        use_pileup = 0
        if 'plot_pileup' in tool:
            use_pileup = 1
        else:
            raise Exception("Error: {} is not suported in tracks. Please report on github if you want this feature.".format(tool))

        command = line_[2:]
        if '-o' in command:
            idx = command.index("-o")
            del command[idx:idx+2]
        if '--output_dir' in command:
            idx = command.index("--output_dir")
            del command[idx:idx+2]
        if '--return_plot' not in command:
            command.append('--return_plot')

        args_tool = plot_pileup.argparser().parse_args(command)
        p, n_plot = plot_pileup.run(args_tool)
        if p is None or n_plot == 0:
            continue
        p.sizing_mode = 'scale_width'
        num_plots.append(n_plot)
        total_plots += n_plot

        if num_tracks > 0 and args.shared_x:
            if len(p.x_range.js_property_callbacks) > 1:
                raise Exception("Error: more than the expected number of callbacks found ({}). please report on github with a minimal dataset to reproduce this error.".format(len(p.x_range.js_property_callbacks)))
            pileup[num_tracks-1].x_range.js_property_callbacks['change:start'] += p.x_range.js_property_callbacks['change:start']
            p.x_range = pileup[num_tracks-1].x_range
            # p.y_range = pileup[0].y_range
        pileup.append(p)
        # output_file(args.output_dir+"/"+str(num_tracks)+".html", title=num_tracks)
        # save(p)
        num_tracks += 1
    if len(pileup) == 0:
        raise Exception("Error: no plots to plot")
    if args.auto_height:
        j = 0
        for p in pileup:
            # print(num_plots[j],(num_plots[j]/total_plots)*PLOT_HEIGHT_SCALE)
            p.height = int((num_plots[j]/total_plots)*PLOT_HEIGHT_SCALE)
            j += 1
    else:
        j = 0
        for p in pileup:
            p.height = int(plot_heights[i])
            j += 1

    if not os.path.exists(args.output_dir):
        os.mkdir(args.output_dir)
    plot_title = "track"
    if args.tag_name:
        plot_title = args.tag_name
    pileup_output_file_name = args.output_dir + "/" + plot_title + ".html"
    pileup_fig = column(pileup, sizing_mode='stretch_both')
    output_file(pileup_output_file_name, title=plot_title)
    save(pileup_fig)
    print(f'output file: {os.path.abspath(pileup_output_file_name)}')

def argparser():
    # parser = argparse.ArgumentParser()
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument('-f', '--file', required=True, help="commands file")
    parser.add_argument('-o', '--output_dir', required=True, help="output dir")
    parser.add_argument('--tag_name', required=False, type=str, default="", help="a tag name to easily identify the plot")
    parser.add_argument('--shared_x', required=False, action='store_true', help="share x-axis so that all the plots move together")
    parser.add_argument('--auto_height', required=False, action='store_true', help="adjust track height automatically")
    return parser

if __name__ == "__main__":
    parser = argparser()
    args = parser.parse_args()
    try:
        run(args)
    except Exception as e:
        print(str(e))
        exit(1)



"""
Signal to seQuence alignment Plot - plot
Hiruna Samarakoon - Garvan Medical Institute
hiruna@unsw.edu.au
"""
import numpy as np
from bokeh.plotting import figure, show, output_file, save
from bokeh.models import ColumnDataSource, LabelSet, CustomJS
from bokeh.colors import RGB
import argparse
import re

# ref_start is always 1based closed
# ref_end is always 1based closed
# start_kmer is always 0based closed
# end_kmer is always 0based open

PLOT_X_RANGE = 750

DEFAULT_NUM_BED_COLS = 3
DEFAULT_BED_ANNOTATION_COLOR = (75, 126, 246)
BED_CHROM, BED_CHROM_START, BED_CHROM_END, BED_NAME, BED_SCORE, BED_STRAND, BED_THICK_START, BED_THICK_END, BED_ITEM_RGB, BED_BLOCK_COUNT, BED_BLOCK_SIZES, BLOCK_STARTS = range(12)


def adjust_bed_for_rna(bed_content, sig_algn_data):
    ref_end = sig_algn_data["ref_end"]
    new_bed_content = []
    for i in bed_content:
        bed_chrom_start = int(i[BED_CHROM_START])
        bed_chrom_end = int(i[BED_CHROM_END])
        i[BED_CHROM_START] = ref_end - bed_chrom_end
        i[BED_CHROM_END] = ref_end - bed_chrom_start
        if i[BED_CHROM_END] > 0:
            new_bed_content.append(i)
    return new_bed_content
def draw_bed_annotation(p, bed_content, sig_algn_data, draw_data, base_limit, track_shift, track_height):
    moves = sig_algn_data["ss"]
    base_index = sig_algn_data["start_kmer"]
    ref_start = sig_algn_data["ref_start"]
    if sig_algn_data["data_is_rna"]:
        ref_start = 1
    ref_end = sig_algn_data["ref_end"]

    if sig_algn_data["data_is_rna"]:
        bed_content = adjust_bed_for_rna(bed_content, sig_algn_data)
    else:
        new_bed_content = []
        for i in bed_content:
            if int(i[BED_CHROM_END]) > base_index+ref_start:
                new_bed_content.append(i)
        bed_content = new_bed_content
    if len(bed_content) == 0:
        return p
    bed_content.sort(key=lambda x: x[BED_CHROM_START])
    annotation_box_details = {'left': [], 'right': [], 'fill_color': []}
    annotation_label = []
    annotation_label_x = []
    location_plot = 0
    x_coordinate = 0
    initial_x_coordinate = x_coordinate
    flag_base_index_bound = 0

    bed_index = 0
    bed_region_start = int(bed_content[bed_index][BED_CHROM_START])
    bed_region_end = int(bed_content[bed_index][BED_CHROM_END])

    flag_region_active = 1
    flag_interval_open = 0
    if len(bed_content[bed_index]) >= BED_ITEM_RGB + 1:
        bed_region_color = eval(bed_content[bed_index][BED_ITEM_RGB])
    else:
        bed_region_color = DEFAULT_BED_ANNOTATION_COLOR
    for i in moves:
        previous_location = location_plot
        previous_x_coordinate = x_coordinate
        if 'D' in i:
            i = re.sub('D', '', i)
            n_samples = int(i)
            prev_loc = previous_location
            prev_x_cord = previous_x_coordinate
            for j in range(0, n_samples):
                prev_loc += draw_data["fixed_base_width"]
                prev_x_cord += draw_data["fixed_base_width"]
                #start add annotation
                if flag_interval_open == 0 and flag_region_active == 1 and bed_region_start+1 == base_index+ref_start:
                    annotation_box_details['left'].append(prev_loc-draw_data["fixed_base_width"])
                    flag_interval_open = 1
                if flag_interval_open == 1 and flag_region_active == 1 and bed_region_end == base_index+ref_start:
                    annotation_box_details['right'].append(prev_loc)
                    annotation_box_details['fill_color'].append(bed_region_color)
                    bed_name = "{}-{} ".format(bed_region_start+1, bed_region_end)
                    if sig_algn_data["data_is_rna"]:
                        bed_name = "{}-{} ".format(ref_end-bed_region_start, ref_end-bed_region_end+1)
                    if len(bed_content[bed_index]) >= BED_NAME + 1:
                        bed_name += bed_content[bed_index][BED_NAME]
                    annotation_label.append(bed_name)
                    annotation_label_x.append((annotation_box_details['left'][bed_index]))
                    flag_interval_open = 0
                    flag_region_active = 0

                if flag_region_active == 0 and bed_region_end == base_index+ref_start and bed_index < len(bed_content)-1:
                    bed_index += 1
                    prev_bed_end = bed_region_end
                    bed_region_start = int(bed_content[bed_index][BED_CHROM_START])
                    bed_region_end = int(bed_content[bed_index][BED_CHROM_END])
                    if prev_bed_end > bed_region_start:
                        raise Exception("Error: overlapping regions found in bed file")
                    if len(bed_content[bed_index]) >= BED_ITEM_RGB + 1:
                        bed_region_color = eval(bed_content[bed_index][BED_ITEM_RGB])
                    else:
                        bed_region_color = DEFAULT_BED_ANNOTATION_COLOR
                    flag_region_active = 1
                #end add annotation
                base_index += 1
                if base_index - sig_algn_data["start_kmer"] == base_limit:
                    flag_base_index_bound = 1
                    break
            if flag_base_index_bound == 1:
                break
            location_plot = prev_loc
            x_coordinate = prev_x_cord

        elif 'I' in i:
            i = re.sub('I', '', i)
            n_samples = int(i)
            location_plot += n_samples
            x_coordinate += n_samples
        else:
            n_samples = int(i)
            if draw_data['fixed_width']:
                location_plot += draw_data["fixed_base_width"]
            else:
                location_plot += n_samples
            x_coordinate += n_samples
            #start add annotation
            if flag_interval_open == 0 and flag_region_active == 1 and bed_region_start+1 <= base_index+ref_start:
                annotation_box_details['left'].append(previous_location)
                flag_interval_open = 1
            if flag_interval_open == 1 and flag_region_active == 1 and bed_region_end == base_index+ref_start:
                annotation_box_details['right'].append(location_plot)
                annotation_box_details['fill_color'].append(bed_region_color)
                bed_name = "{}-{} ".format(bed_region_start+1, bed_region_end)
                if sig_algn_data["data_is_rna"]:
                    bed_name = "{}-{} ".format(ref_end-bed_region_start, ref_end-bed_region_end+1)
                if len(bed_content[bed_index]) >= BED_NAME + 1:
                    bed_name += bed_content[bed_index][BED_NAME]
                annotation_label.append(bed_name)
                annotation_label_x.append((annotation_box_details['left'][bed_index]))
                flag_interval_open = 0
                flag_region_active = 0

            if flag_region_active == 0 and bed_region_end == base_index+ref_start and bed_index < len(bed_content)-1:
                bed_index += 1
                prev_bed_end = bed_region_end
                bed_region_start = int(bed_content[bed_index][BED_CHROM_START])
                bed_region_end = int(bed_content[bed_index][BED_CHROM_END])
                if prev_bed_end > bed_region_start:
                    raise Exception("Error: overlapping regions found in bed file")
                if len(bed_content[bed_index]) >= BED_ITEM_RGB + 1:
                    bed_region_color = eval(bed_content[bed_index][BED_ITEM_RGB])
                else:
                    bed_region_color = DEFAULT_BED_ANNOTATION_COLOR
                flag_region_active = 1
            #end add annotation
            base_index += 1

        if base_index - sig_algn_data["start_kmer"] == base_limit:
            break
        if x_coordinate - initial_x_coordinate > draw_data["sig_plot_limit"]:
            break

    if flag_interval_open == 1 and flag_region_active == 1:
        if len(annotation_box_details['left']) != len(annotation_box_details['right']) + 1:
            raise Exception("Error: bed annotation interval dimensions are incorrect. Please report")
        annotation_box_details['right'].append(location_plot)
        annotation_box_details['fill_color'].append(bed_region_color)
        bed_name = "{}- ".format(bed_region_start+1)
        if sig_algn_data["data_is_rna"]:
            bed_name = "{}- ".format(ref_end-bed_region_start)
        if len(bed_content[bed_index]) >= BED_NAME + 1:
            bed_name += bed_content[bed_index][BED_NAME]
        annotation_label.append(bed_name)
        annotation_label_x.append((annotation_box_details['left'][bed_index]))
    # print(annotation_box_details)
    p.quad(top=draw_data['y_max']+track_shift+track_height, bottom=draw_data['y_max']+track_shift, left=annotation_box_details['left'], right=annotation_box_details['right'], color=annotation_box_details['fill_color'], alpha=0.75)

    # print(annotation_label)
    if draw_data["bed_labels"]:
        bed_annotation = ColumnDataSource(data=dict(base_x=annotation_label_x, base_label=annotation_label))
        bed_annotation_labels = LabelSet(x='base_x', y=draw_data['y_max']+track_shift, text='base_label', source=bed_annotation, text_font_size="7pt")
        p.add_layout(bed_annotation_labels)

        x_callback = CustomJS(args=dict(bed_annotation_labels=bed_annotation_labels, init_font_size=bed_annotation_labels.text_font_size[:-2], init_xrange=PLOT_X_RANGE), code="""
        let xzoom = (init_font_size * init_xrange) / (cb_obj.end - cb_obj.start);
        bed_annotation_labels['text_font_size'] = String(xzoom) + 'pt';
        """)
        p.x_range.js_on_change('start', x_callback)

    return p
def plot_bed_annotation(p, ref_id, bed_dic, sig_algn_data, draw_data, base_limit):
    if ref_id in bed_dic:
        plot_height = draw_data['y_max'] - draw_data['y_min']
        track_height = plot_height/5
        # print("track_height: {}", track_height)
        track_count = 0
        for key in bed_dic[ref_id]:
            track_count += 1
            draw_bed_annotation(p, bed_dic[ref_id][key], sig_algn_data, draw_data, base_limit, track_height*track_count, track_height)
    return p
def create_bed_dic(args):
    bed_content = []
    bed_num_cols = DEFAULT_NUM_BED_COLS
    count = 0
    with open(args.bed)as f:
        for line in f:
            bed_list = line.strip().split()
            if len(bed_list) < DEFAULT_NUM_BED_COLS:
                raise Exception("Error: minimum {} columns required in a bed file", bed_num_cols)
            if count == 0:
                bed_num_cols = len(bed_list)
            if len(bed_list) != bed_num_cols:
                raise Exception("Error: Number of columns vary for different bed records")
            count += 1
            bed_content.append(bed_list)
    # print(bed_content)
    bed_dic = {}
    bed_chrom_set = set()
    for i in range(0, len(bed_content)):
        bed_chrom_set.add(bed_content[i][BED_CHROM])
    for i in bed_chrom_set:
        bed_dic[i] = {}
    if len(bed_content) > 0:
        if bed_num_cols >= BED_NAME + 1:
            for i in range(0, len(bed_content)):
                if bed_content[i][BED_NAME] not in bed_dic[bed_content[i][BED_CHROM]]:
                    bed_dic[bed_content[i][BED_CHROM]][bed_content[i][BED_NAME]] = []
                bed_dic[bed_content[i][BED_CHROM]][bed_content[i][BED_NAME]].append(bed_content[i])
        else:
            for i in range(0, len(bed_content)):
                if "DEFAULT" not in bed_dic[bed_content[i][BED_CHROM]]:
                    bed_dic[bed_content[i][BED_CHROM]]["DEFAULT"] = []
                bed_dic[bed_content[i][BED_CHROM]]["DEFAULT"].append(bed_content[i])
    return bed_dic





import os
import argparse
from readpaf import parse_paf
import pysam

BAM_CMATCH, BAM_CINS, BAM_CDEL, BAM_CREF_SKIP, BAM_CSOFT_CLIP, BAM_CHARD_CLIP, BAM_CPAD, BAM_CEQUAL, BAM_CDIFF, BAM_CBACK = range(10)

def run(args):

    print("input bam: " + args.bam)
    print("input paf: " + args.paf)
    print("output file: " + args.output)
    print("INFO: supplementary alignments will be skipped.")
    # print("INFO: move array of a reverse complemented alignment will be treated after reversing.")

    samfile = pysam.AlignmentFile(args.bam, mode='r')
    fout = ""
    if args.output[-4:] == ".bam":
        fout = pysam.AlignmentFile(args.output, "wb", template=samfile)
    elif args.output[-4:] == ".sam":
        fout = pysam.AlignmentFile(args.output, "w", template=samfile)
    elif args.output[-4:] == ".paf":
        if args.c:
            fout = open(args.output, "w")
        else:
            raise Exception("Error: please provide argument '-c' to write in PAF format")
    else:
        raise Exception("Error: please provide the output file with correct extension")

    # inefficient
    paf_file = open(args.paf, "r")
    paf_dic = {}
    for record in parse_paf(paf_file):
        paf_dic[record.query_name] = record

    processed_sam_record_count = 0
    for sam_record in samfile:
        if sam_record.is_unmapped or sam_record.is_supplementary or sam_record.is_secondary:
            continue
        # if sam_record.query_name != "80922061-df02-48ad-bd67-65e5adcc78f5" and sam_record.query_name != "a0d0047d-64d1-4cfd-9792-8c27ad9a8c06":
        #     continue
        sam_read_id = sam_record.query_name
        if sam_read_id not in paf_dic:
            raise Exception("Error: associated paf record is missing for the read id: {}".format(sam_read_id))
        paf_read_id = paf_dic[sam_read_id].query_name
        if paf_read_id != sam_read_id:
            raise Exception("Error: sam and paf read ids do not match")

        data_is_rna = False
        if paf_dic[sam_read_id].target_start > paf_dic[sam_read_id].target_end:  # if RNA start_kmer>end_kmer in paf
            data_is_rna = True
            if not args.rna:
                print("Info: data is detected as RNA")
                raise Exception("Error: data is not specified as RNA. Please provide the argument --rna ")
        if not data_is_rna and args.rna:
            raise Exception("Error: data is not not detected as RNA but the user specified as RNA. Please remove the argument --rna and check dataset")

        # print(sam_read_id)
        # print("sam_read.pos: " + str(sam_read.pos+1))
        # print(sam_read.cigarstring)
        cigar_t = sam_record.cigartuples
        if cigar_t is None:
            raise Exception("Error: cigartuples for sam record {} is an empty object".format(sam_read_id))
        # print(cigar_t)

        moves_string = paf_dic[sam_read_id].tags['ss'][2].rstrip(',').split(',')
        raw_start = paf_dic[sam_read_id].query_start
        raw_end = paf_dic[sam_read_id].query_end
        # print("{}\t{}\t{}\t{}\t{}\t{}".format(raw_start, raw_end, kmer_start, kmer_end, len(sam_record.seq), abs(kmer_end-kmer_start)))
        # print(sam_record.cigarstring)
        # print(sam_record.pos+1)
        strand_dir = "+"
        if sam_record.is_reverse:
            cigar_t.reverse()
            strand_dir = "-"
        if data_is_rna:
            cigar_t.reverse()
            # strand_dir = "RNA"

        idx = 0
        ss_string = ""
        count_bases = 0
        count_bases_seq = 0
        # print(len(moves_string))
        # print(paf_dic[sam_read_id].target_end)
        # print(paf_dic[sam_read_id].target_start)
        # print(len(sam_record.query_sequence))
        # if len(sam_record.query_sequence) != len(moves_string):
        #     raise Exception("Error: the sequence length does not match the number of moves")
        len_moves = len(moves_string)
        op_count = 0
        total_cig_count_seq = 0
        for a in cigar_t:
            cig_op = a[0]
            cig_count = a[1]
            if cig_op == BAM_CMATCH:
                if total_cig_count_seq + cig_count > len_moves:
                    cig_count = len_moves - total_cig_count_seq
                total_cig_count_seq += cig_count
                for i in range(0, cig_count):
                    ss_string = ss_string + moves_string[idx] + ","
                    idx = idx + 1
                count_bases += cig_count
                count_bases_seq += cig_count
            elif cig_op == BAM_CDEL:
                ss_string = ss_string + str(cig_count) + "D"
                count_bases += cig_count
                # print(str(cig_count) + " D " + str(int(sam_read.pos) + 1 + count_bases))
            elif cig_op == BAM_CINS:
                if total_cig_count_seq + cig_count > len_moves:
                    cig_count = len_moves - total_cig_count_seq
                total_cig_count_seq += cig_count
                signal_skip = 0
                for i in range(0, cig_count):
                    signal_skip = signal_skip + int(moves_string[idx])
                    idx = idx + 1
                ss_string = ss_string + str(signal_skip) + "I"
                count_bases_seq += cig_count
                # print(str(signal_skip) + " I BAM_CINS " + str(int(sam_read.pos) + 1 + count_bases))
            elif cig_op == BAM_CSOFT_CLIP:
                if total_cig_count_seq + cig_count > len_moves:
                    cig_count = len_moves - total_cig_count_seq
                total_cig_count_seq += cig_count
                signal_skip = 0
                for i in range(0, cig_count):
                    # print(str(idx) + " " + moves_string[idx])
                    signal_skip = signal_skip + int(moves_string[idx])
                    idx = idx + 1
                if op_count == 0:
                    raw_start += signal_skip
                else:
                    raw_end -= signal_skip
                # ss_string = ss_string + str(signal_skip) + "I"
                # print(str(signal_skip) + " I BAM_CSOFT_CLIP " + str(int(sam_read.pos) + 1 + count_bases))
            elif cig_op == BAM_CHARD_CLIP:
                continue
            elif cig_op == BAM_CREF_SKIP:
                continue
            else:
                raise Exception("Error: cigar operation [" + str(cig_op) + "]is not handled yet found in read : " + sam_read_id)
            op_count += 1

        kmer_start = sam_record.reference_start
        kmer_end = sam_record.reference_start + count_bases
        if args.c:
            if data_is_rna:
                paf_record_str = "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\tss:Z:{}\n".format(paf_dic[sam_read_id].query_name, paf_dic[sam_read_id].query_length, raw_start, raw_end, strand_dir, sam_record.reference_name, count_bases, kmer_end, kmer_start, count_bases, count_bases, "255", ss_string)
            else:
                paf_record_str = "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\tss:Z:{}\n".format(paf_dic[sam_read_id].query_name, paf_dic[sam_read_id].query_length, raw_start, raw_end, strand_dir, sam_record.reference_name, count_bases, kmer_start, kmer_end, count_bases, count_bases, "255", ss_string)
            fout.write(paf_record_str)
        else:
            #start_raw, end_raw, start_kmer and end_kmer,
            if data_is_rna:
                si_string = str(raw_start) + "," + str(raw_end) + "," + str(kmer_end) + "," + str(kmer_start)
            else:
                si_string = str(raw_start) + "," + str(raw_end) + "," + str(kmer_start) + "," + str(kmer_end)

            # print("{}\t{}\t{}\t{}\t{}\t{}".format(strand_dir, sam_read_id, si_string, count_bases, count_bases_seq, abs(kmer_end-kmer_start)))
            sam_record.set_tag("ss", ss_string, "Z")
            sam_record.set_tag("si", si_string, "Z")
            fout.write(sam_record)
        processed_sam_record_count += 1

    print("processed_sam_record_count: " + str(processed_sam_record_count))

    paf_file.close()
    samfile.close()
    fout.close()

def argparser():
    # parser = argparse.ArgumentParser()
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument('-p', '--paf', required=True, help="input read-signal alignment .paf file")
    parser.add_argument('-b', '--bam', required=True, help="input read-reference alignment SAM/BAM file")
    parser.add_argument('-c', action='store_true', help="write move table in paf format")
    parser.add_argument('-o', '--output', required=True, help="output reference-signal SAM/BAM/PAF file")
    parser.add_argument('--rna', required=False, action='store_true', help="specify for RNA reads")

    return parser

if __name__ == "__main__":
    parser = argparser()
    args = parser.parse_args()
    try:
        run(args)
    except Exception as e:
        print(str(e))
        exit(1)



"""
Signal to seQuence alignment Plot - plot
Hiruna Samarakoon - Garvan Medical Institute
hiruna@unsw.edu.au
"""
from bokeh.plotting import figure, show, output_file, save
from bokeh.models import BoxAnnotation, HoverTool, WheelZoomTool, ColumnDataSource, Label, LabelSet, Segment, Toggle, Range1d, FreehandDrawTool
import re
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import seaborn as sns

PLOT_X_RANGE = 300
PLOT_HEIGHT = 600

def adjust_before_plotting(ref_seq_len, signal_tuple, region_tuple, sig_algn_data, fasta_seq):
    if sig_algn_data["data_is_rna"]:
        ref_region_start_diff = region_tuple[1] - region_tuple[3]
    else:
        ref_region_start_diff = region_tuple[2] + 1 - region_tuple[0]
    ref_region_end_diff = region_tuple[2] + ref_seq_len - region_tuple[1]
    # print("ref_region_start_diff: " + str(ref_region_start_diff))
    moves = sig_algn_data['ss']
    if ref_region_start_diff < 0:
        # we have to remove part of the sequence until we get to the ref start
        count_bases = 0
        eat_signal = 0
        count_moves = 0
        prev_move = None
        updated_move = []
        for i in moves:
            if count_bases == abs(ref_region_start_diff):
                break
            if count_bases > abs(ref_region_start_diff):
                if not prev_move.find('D'):
                    raise Exception("Error: a deletion move was expected. incorrect implementation. Please report with a minimal reproducible test")
                updated_move = ["{}D".format(count_bases-abs(ref_region_start_diff))]
                break
            count_moves += 1

            if 'D' in i:
                i = re.sub('D', '', i)
                count_bases += int(i)
                # print(i+" D "+str(int(sam_record.reference_start) + 1 + count_bases))
            elif 'I' in i:
                i = re.sub('I', '', i)
                eat_signal += int(i)
                # print(i+" I "+str(int(sam_record.reference_start) + 1 + count_bases))
            else:
                eat_signal += int(i)
                count_bases += 1
            prev_move = i
        moves = updated_move + moves[count_moves:]
        x = signal_tuple[0][:-eat_signal]
        x_real = signal_tuple[1][eat_signal:]
        y = signal_tuple[2][eat_signal:]
        signal_tuple = (x, x_real, y)
        sig_algn_data['ss'] = moves
    return signal_tuple, region_tuple, sig_algn_data, fasta_seq
def create_figure(args, plot_mode):
    p_defualt = None
    if plot_mode == 0:
        y_axis_label = "signal value (raw)"
        if args.no_pa:
            y_axis_label = "signal value (pA)"
        tools_to_show = 'hover,box_zoom,pan,reset,save,wheel_zoom,zoom_in,zoom_out'
        p_default = figure(x_axis_label='signal index',
                   y_axis_label=y_axis_label,
                   sizing_mode="stretch_width",
                   height=PLOT_HEIGHT,
                   output_backend="webgl",
                   x_range=(0, PLOT_X_RANGE),
                   tools=tools_to_show,
                   toolbar_location="below")
        # tooltips=tool_tips)
        p_default.select(dict(type=WheelZoomTool)).maintain_focus = False
        p_default.toolbar.active_scroll = p_default.select_one(WheelZoomTool)
        p_default.toolbar.logo = None
    elif plot_mode == 1:
        tools_to_show = 'hover,box_zoom,pan,save,wheel_zoom,reset,zoom_in,zoom_out'
        p_default = figure(output_backend="webgl",
                    sizing_mode="stretch_both",
                   # sizing_mode="scale_width",
                   # sizing_mode="scale_height",
                   # height=PLOT_HEIGHT,
                   x_range=(0, PLOT_X_RANGE),
                   tools=tools_to_show)
        # tooltips=tool_tips)
        # p.yaxis.visible = False
        p_default.select(dict(type=WheelZoomTool)).maintain_focus = False
        p_default.toolbar.active_scroll = p_default.select_one(WheelZoomTool)
        p_default.toolbar.logo = None
    return p_default
def scale_signal(y, sig_scale):
    if sig_scale == "medmad":
        arr = np.ma.array(y).compressed()
        read_median = np.median(arr)
        if read_median == np.nan:
            raise Exception("Error: calculated median is NaN")
        mad = np.median(np.abs(arr - read_median))
        if mad == np.nan:
            raise Exception("Error: calculated mad is NaN")
        read_mad = mad * 1.4826
        if read_mad < 1.0:
            read_mad = 1.0
        y = (y - read_mad) / read_mad
    elif sig_scale == "znorm":
        # zsig = sklearn.preprocessing.scale(y, axis=0, with_mean=True, with_std=True, copy=True)
        # Calculate the z-score from scratch
        y = (y - np.mean(y)) / np.std(y)
    elif not sig_scale == "":
        raise Exception("Error: given --sig_scale method: {} is not supported".format(sig_scale))
    return y
profile_dic_base_shift = {
        "kmer_model_dna_r9.4.1_450bps_5_mer": [-2, -2],
        "kmer_model_dna_r9.4.1_450bps_6_mer": [-2, -3],
        "kmer_model_rna_r9.4.1_70bps_5_mer": [-3, -1],
        "kmer_model_dna_r10.4.1_e8.2_400bps_9_mer": [-6, -2],
        "guppy_dna_r9.4.1_450bps_fast_prom": [0, 0],
        "guppy_dna_r9.4.1_450bps_hac_prom": [0, 0],
        "guppy_dna_r9.4.1_450bps_sup_prom": [0, 0],
        "guppy_dna_r10.4.1_e8.2_400bps_fast": [0, 0],
        "guppy_dna_r10.4.1_e8.2_400bps_hac": [0, 0],
        "guppy_dna_r10.4.1_e8.2_400bps_sup": [0, 0]}
def list_profiles_base_shift():
    # print(profile_dic)
    print("{}\t{}\t{}".format("name", "base_shift_forward", "base_shift_reverse"))
    for profile in profile_dic_base_shift:
        print("{}\t{}\t{}".format(profile, profile_dic_base_shift[profile][0], profile_dic_base_shift[profile][1]))
    print("If the profile you wanted is not listed here, please refer calculate_offsets.md on github to learn how to generate base shift values for your new data.")
def search_for_profile_base_shift(profile):
    if profile in profile_dic_base_shift:
        return profile_dic_base_shift[profile]
    else:
        raise Exception("Error: specified profile ({}) is not found. Please run reform with -k 1 -s 0. Then run calculate_offsets.py and rerun reform with the recommended kmer_length and sig_move_offset.".format(profile))



"""
Signal to seQuence alignment Plot - plot
Hiruna Samarakoon - Garvan Medical Institute
hiruna@unsw.edu.au
"""
import numpy as np
from bokeh.plotting import figure, show, output_file, save
from bokeh.models import HoverTool, WheelZoomTool, ColumnDataSource, Label, LabelSet, Segment, Arrow, NormalHead, FreehandDrawTool, Range1d, CustomJS
from bokeh.layouts import column
from bokeh.colors import RGB
import pyslow5
import copy
import argparse
import re
import logging
from readpaf import parse_paf
from sys import stdin
from pyfaidx import Fasta
from pyfastx import Fastq
import os
import pysam
import math
from src import bed_annotation
from src import plot_utils
import cProfile, pstats, io
# ref_start is always 1based closed
# ref_end is always 1based closed
# start_kmer is always 0based closed
# end_kmer is always 0based open

FIXED_BASE_WIDTH = 10
FIXED_INSERTION_WIDTH = 10
BASE_LIMIT = 1000
SIG_PLOT_LENGTH = 20000
DEFAULT_STRIDE = 5
PLOT_X_RANGE = 750
# PLOT_HEIGHT = 900
PLOT_Y_MARGIN = 1
SUBPLOT_X = -100
PLOT_BASE_SHIFT = 0
PLOT_X_PADDING = 200

BAM_CMATCH, BAM_CINS, BAM_CDEL, BAM_CREF_SKIP, BAM_CSOFT_CLIP, BAM_CHARD_CLIP, BAM_CPAD, BAM_CEQUAL, BAM_CDIFF, BAM_CBACK = range(10)
READ_ID, LEN_RAW_SIGNAL, START_RAW, END_RAW, STRAND, SEQUENCE_ID, LEN_KMER, START_KMER, END_KMER, MATCHES, LEN_KMER, MAPQ = range(12)
SI_START_RAW, SI_END_RAW, SI_START_KMER, SI_END_KMER = range(4)

def plot_function_fixed_width_pileup(read_id, signal_tuple, sig_algn_data, fasta_sequence, base_limit, draw_data, p, num_plots, y_shift, y_max, y_min):
    x = signal_tuple[0]
    x_real = signal_tuple[1]
    y = signal_tuple[2]

    label_position = y_min

    # base_color_map = {'A': 'limegreen', 'C': 'blue', 'T': 'red', 'G': 'orange', 'U': 'red', 'N': 'lavender'}
    base_color_map = {'A': '#d6f5d6', 'C': '#ccccff', 'T': '#ffcccc', 'G': '#ffedcc', 'U': '#ffcccc', 'N': '#fafafe'}
    base_x = []
    base_y = []
    base_label = []
    base_label_colors = []

    move_x = []
    move_y = []
    move_label = []
    move_label_colors = []

    sample_label_colors_match = []
    sample_label_colors_insert = []
    location_plot = 0

    x_coordinate = 0
    initial_x_coordinate = x_coordinate

    # draw moves
    moves = sig_algn_data["ss"]
    base_index = sig_algn_data["start_kmer"]
    num_Is = 0
    num_Ds = 0
    fixed_width_x = [0.0]
    line_segment_x = []

    base_box_details = {'left': [], 'right': [], 'fill_color': []}
    flag_base_index_bound = 0
    num_samples_in_insertion = 0

    for i in moves:
        previous_location = location_plot
        previous_x_coordinate = x_coordinate
        if 'D' in i:
            i = re.sub('D', '', i)
            n_samples = int(i)

            prev_loc = previous_location
            prev_x_cord = previous_x_coordinate
            for j in range(0, n_samples):
                base = fasta_sequence[base_index]
                if num_plots == -1 or draw_data['overlap_only']:
                    base_label_colors.append('black')
                    base_box_details['left'].append(prev_loc)
                    base_box_details['right'].append(prev_loc + draw_data["fixed_base_width"])
                    base_box_details['fill_color'].append(base_color_map[base])
                else:
                    base_label_colors.append('red')
                    base_box_details['left'].append(prev_loc)
                    base_box_details['right'].append(prev_loc + draw_data["fixed_base_width"])
                    base_box_details['fill_color'].append('white')

                base_x.append(prev_loc)
                base_y.append(label_position+y_shift)
                label = str(base) + "\n" + str(base_index + 1)
                base_label.append(label)

                prev_loc += draw_data["fixed_base_width"]
                prev_x_cord += draw_data["fixed_base_width"]
                base_index += 1
                num_Ds += 1
                if base_index - sig_algn_data["start_kmer"] == base_limit:
                    flag_base_index_bound = 1
                    break

            if flag_base_index_bound == 1:
                break
            location_plot = prev_loc
            x_coordinate = prev_x_cord

            fixed_width_x = fixed_width_x + list(range(int(fixed_width_x[-1]) + 1, int(fixed_width_x[-1]) + 1 + n_samples * draw_data["fixed_base_width"]))
            y_add = np.concatenate((y[:previous_x_coordinate], [np.nan] * n_samples * draw_data["fixed_base_width"]), axis=0)
            y = np.concatenate((y_add, y[previous_x_coordinate:]), axis=0)
            x_add = np.concatenate((x_real[:previous_x_coordinate], [x_real[previous_x_coordinate]] * n_samples * draw_data["fixed_base_width"]), axis=0)
            x_real = np.concatenate((x_add, x_real[previous_x_coordinate:]), axis=0)

            for j in range(0, n_samples * draw_data["fixed_base_width"]):
                sample_label_colors_match.append('None')
                sample_label_colors_insert.append('None')

        elif 'I' in i:
            i = re.sub('I', '', i)
            n_samples = int(i)
            # prev_x_value = fixed_width_x[-1]
            # for s in range(0, n_samples-1):
            #     fixed_width_x.append(fixed_width_x[-1] + draw_data["fixed_base_width"]/n_samples)
            # fixed_width_x.append(prev_x_value + draw_data["fixed_base_width"])
            num_samples_in_insertion = n_samples
            num_Is += n_samples
            # location_plot += draw_data["fixed_base_width"]
            # x_coordinate += n_samples
            # line_segment_x.append(location_plot)
            for j in range(0, n_samples):
                sample_label_colors_match.append('None')
                sample_label_colors_insert.append('purple')
        else:
            n_samples = int(i) + num_samples_in_insertion
            prev_x_value = fixed_width_x[-1]
            for s in range(0, n_samples - 1):
                fixed_width_x.append(fixed_width_x[-1] + draw_data["fixed_base_width"] / n_samples)
            fixed_width_x.append(prev_x_value + draw_data["fixed_base_width"])

            location_plot += draw_data["fixed_base_width"]
            x_coordinate += n_samples

            base = fasta_sequence[base_index]
            base_box_details['left'].append(previous_location)
            base_box_details['right'].append(location_plot)
            base_box_details['fill_color'].append(base_color_map[base])
            line_segment_x.append(location_plot)
            if num_plots == -1 or draw_data['overlap_only']:
                base_x.append(previous_location)
                base_y.append(label_position + y_shift)
                label = str(base) + "\n" + str(base_index + 1)
                base_label.append(label)
                base_label_colors.append('black')
            else:
                if num_samples_in_insertion > 0:
                    move_x.append(previous_location)
                    move_y.append(label_position + y_shift)
                    move_label.append(str(num_samples_in_insertion))
                    move_label_colors.append('purple')
                move_x.append(previous_location+draw_data["fixed_base_width"]/3)
                move_y.append(label_position + y_shift)
                move_label.append(str(n_samples))
                move_label_colors.append('black')

            for j in range(0, n_samples - num_samples_in_insertion):
                sample_label_colors_match.append('red')
                sample_label_colors_insert.append('None')
            base_index += 1
            num_samples_in_insertion = 0

        if base_index - sig_algn_data["start_kmer"] == base_limit:
            break
        if x_coordinate - initial_x_coordinate > draw_data["sig_plot_limit"]:
            break

    line_segment_source = ColumnDataSource(dict(x=line_segment_x, x1=line_segment_x, y=[y_min+y_shift] * len(line_segment_x), y1=[y_max+y_shift] * len(line_segment_x)))
    glyph = Segment(x0="x", y0="y", x1="x1", y1="y1", line_color="#8b4513", line_width=1, line_alpha=0.5)

    base_annotation = ColumnDataSource(data=dict(base_x=base_x, base_y=base_y, base_label=base_label, colors=base_label_colors))
    base_annotation_labels = LabelSet(x='base_x', y='base_y', text='base_label', x_offset=5, y_offset=5, source=base_annotation, text_font_size="7pt", text_color='colors')

    move_annotation = ColumnDataSource(data=dict(move_x=move_x, move_y=move_y, move_label=move_label, colors=move_label_colors))
    move_annotation_labels = LabelSet(x='move_x', y='move_y', text='move_label', x_offset=5, y_offset=5, source=move_annotation, text_font_size="5pt", text_color='colors', visible=True)

    fixed_width_x = fixed_width_x[1:]
    source = ColumnDataSource(data=dict(x=fixed_width_x[:x_coordinate], y=y[:x_coordinate]+y_shift, x_real=x_real[:x_coordinate], y_real=y[:x_coordinate]))

    if (draw_data['overlap_only'] and num_plots == 0) or not draw_data['overlap_only']:
        p.quad(top=y_max+y_shift, bottom=y_min+y_shift, left=base_box_details['left'], right=base_box_details['right'], color=base_box_details['fill_color'], alpha=0.75)
        p.add_glyph(line_segment_source, glyph)
        p.add_layout(base_annotation_labels)
        if draw_data["plot_num_samples"]:
            p.add_layout(move_annotation_labels)
            x_callback_move_annotation = CustomJS(args=dict(move_annotation_labels=move_annotation_labels, init_font_size=move_annotation_labels.text_font_size[:-2], init_xrange=PLOT_X_RANGE), code="""
            let xzoom = (init_font_size * init_xrange) / (cb_obj.end - cb_obj.start);
            move_annotation_labels['text_font_size'] = String(xzoom) + 'pt';
            """)
            p.x_range.js_on_change('start', x_callback_move_annotation)

    if num_plots != -1:
        if num_plots == 0:
            sample_label_colors_insert[0] = 'purple'
            sample_label_colors_match[0] = 'red'
        sample_labels_match = p.circle(fixed_width_x[:x_coordinate], y[:x_coordinate]+y_shift, radius=draw_data["point_size"], color=sample_label_colors_match, alpha=0.5, legend_label='match', visible=False)
        sample_labels_insert = p.circle(fixed_width_x[:x_coordinate], y[:x_coordinate]+y_shift, radius=draw_data["point_size"], color=sample_label_colors_insert, alpha=0.5, legend_label='insertion', visible=False)

    if not draw_data['overlap_only']:
        if num_plots != -1:
            p.line('x', 'y', name="sig_plot_line", line_width=2, source=source)
    if not draw_data['no_overlap']:
        leg_lable = "read_" + str(num_plots+1)
        if num_plots != -1:
            p.line('x', 'y_real', line_width=2, source=source, legend_label=leg_lable)

    # show the tooltip
    hover = p.select(dict(type=HoverTool))
    hover.renderers = p.select(name="sig_plot_line")
    hover.tooltips = [("x", "@x_real"), ("y", "@y_real")]
    hover.mode = 'mouse'

    signal_region = ""
    indels = f'deletions(bases): {num_Ds} insertions(samples): {num_Is}'
    signal_region = f'[{int(x_real[0])}-{int(x_real[x_coordinate - 1])}]'
    y_plot = y[:x_coordinate]+y_shift
    y_median = np.nanmedian(y_plot)
    y_max = np.nanmax(y_plot)
    y_min = np.nanmin(y_plot)
    arrow = Arrow(end=NormalHead(fill_color="orange", size=10), x_start=-2, y_start=y_median, x_end=-1, y_end=y_median)
    subplot_labels = None
    if num_plots != -1 and not draw_data['overlap_only']:
        sub_plot_y_shift = (y_max - y_min)/6
        source_subplot_labels = ColumnDataSource(data=dict(x=[SUBPLOT_X, SUBPLOT_X, SUBPLOT_X, SUBPLOT_X], y=[y_median+sub_plot_y_shift*1, y_median, y_median-sub_plot_y_shift*1, y_median-sub_plot_y_shift*2], tags=[signal_region, indels, read_id, "read: "+str(num_plots+1)]))
        subplot_labels = LabelSet(x='x', y='y', text='tags', text_font_size="7pt", x_offset=5, y_offset=5, source=source_subplot_labels)
        p.add_layout(subplot_labels)
        p.add_layout(arrow)
    elif num_plots == -1 or (draw_data['overlap_only'] and num_plots == 0):
        source_subplot_labels = ColumnDataSource(data=dict(x=[SUBPLOT_X], y=[y_median], tags=["overlap"]))
        subplot_labels = LabelSet(x='x', y='y', text='tags', text_font_size="7pt", x_offset=5, y_offset=5, source=source_subplot_labels)
        p.add_layout(subplot_labels)
        p.add_layout(arrow)
    if subplot_labels:
        x_callback_subplot_labels = CustomJS(args=dict(subplot_labels=subplot_labels, init_font_size=subplot_labels.text_font_size[:-2], init_xrange=PLOT_X_RANGE), code="""
        let xzoom = (init_font_size * init_xrange) / (cb_obj.end - cb_obj.start);
        subplot_labels['text_font_size'] = String(xzoom) + 'pt';
        """)
        p.x_range.js_on_change('start', x_callback_subplot_labels)
    
    x_callback_base_annotation = CustomJS(args=dict(base_annotation_labels=base_annotation_labels, init_font_size=base_annotation_labels.text_font_size[:-2], init_xrange=PLOT_X_RANGE), code="""
    let xzoom = (init_font_size * init_xrange) / (cb_obj.end - cb_obj.start);
    base_annotation_labels['text_font_size'] = String(xzoom) + 'pt';
    """)
    p.x_range.js_on_change('start', x_callback_base_annotation)

    return p, location_plot, base_index
def run(args):
    if args.list_profile:
        plot_utils.list_profiles_base_shift()
        return
    else:
        if args.file == "":
            raise Exception("Error: the following argument is required: -f/--file")
        if args.slow5 == "":
            raise Exception("Error: the following argument is required: -s/--slow5")
        if args.alignment == "":
            raise Exception("Error: the following argument is required: -a/--alignment")
        if args.region == "":
            raise Exception("Error: the following argument is required: --region")
        if not args.return_plot and args.output_dir == "":
            raise Exception("Error: one of the arguments -o/--output_dir --return_plot is required")

    if args.cprofile:
        pr = cProfile.Profile()
        pr.enable()



    use_fasta = 0
    if args.file:
        print(f'sequence file: {args.file}')
        if args.file[-6:] == ".fastq" or args.file[-6:] == ".fq.gz" or args.file[-3:] == ".fq":
            use_fasta = 0
        elif args.file[-6:] == ".fasta" or args.file[-3:] == ".fa":
            use_fasta = 1
        else:
            raise Exception("Error: please provide the sequence file with correct extension")

    max_location_plot = 0
    max_base_index = 0
    plot_sig_ref_flag = 0
    use_paf = 0
    if args.alignment:
        print(f'alignment file: {args.alignment}')
        alignment_extension = args.alignment[-4:]
        if alignment_extension == ".bam" or alignment_extension == ".sam":
            use_paf = 0
            plot_sig_ref_flag = 1
        elif alignment_extension == ".paf":
            raise Exception("Error: For signal to reference alignment the paf file must be bgzip compressed and tabix indexed")
        elif args.alignment[-7:] == ".paf.gz":
            use_paf = 1
            plot_sig_ref_flag = 1
            index_file = args.alignment + ".tbi"
            if not os.path.exists(index_file):
                raise Exception("Error: please provide a bgzip compressed and tabix indexed paf file to extract the regions")
        else:
            raise Exception("Error: please provide the alignment file with correct extension")

    if plot_sig_ref_flag == 0:
        raise Exception("Error: If you are trying to plot just reads please use plot.py. Otherwise provide the correct signal-reference alignment.")

    if use_paf == 0 and use_fasta == 0:
        print("please provide a .fasta or .fa file when using SAM/BAM")

    # if not args.fixed_width:
    #     raise Exception("Error: pileup works only with fixed base width. Provide the argument --fixed_width")

    if args.base_limit:
        base_limit = args.base_limit
    else:
        base_limit = BASE_LIMIT
    print(f'signal file: {args.slow5}')

    if args.read_id != "":
        args.plot_limit = 1
    if args.read_list != "":
        print(f'read_id list file: {args.read_list}')
        read_id_list = list(line.strip() for line in open(args.read_list))

    bed_dic = {}
    if args.bed:
        print(f'bed file: {args.bed}')
        bed_dic = bed_annotation.create_bed_dic(args)
    if args.plot_reverse:
        print("Info: reads mapped to the reverse strand will be plotted")
    if args.return_plot:
        print("Info: plots will be returned without saving")
    else:
        if not os.path.exists(args.output_dir):
            os.mkdir(args.output_dir)
    # if not args.loose_bound:
    #     print("Warning: the alignments that are not completely within the specified region will be skipped. Please specify --loose_bound to keep")

    # open signal file
    s5 = pyslow5.Open(args.slow5, 'r')
    num_plots = 0
    indt = "\t\t\t\t\t\t\t\t"
    draw_data = {}
    draw_data["point_size"] = args.point_size
    draw_data["sig_plot_limit"] = args.sig_plot_limit
    draw_data["fixed_base_width"] = args.base_width
    draw_data["plot_y_margin"] = PLOT_Y_MARGIN
    draw_data["no_overlap"] = args.no_overlap
    draw_data["overlap_only"] = args.overlap_only
    draw_data["plot_num_samples"] = args.plot_num_samples
    draw_data["bed_labels"] = args.print_bed_labels

    if args.profile == "":
        draw_data["base_shift"] = args.base_shift
    else:
        if args.plot_reverse:
            draw_data["base_shift"] = plot_utils.search_for_profile_base_shift(args.profile)[1]
        else:
            draw_data["base_shift"] = plot_utils.search_for_profile_base_shift(args.profile)[0]

    kmer_correction = 0
    if args.profile != "":
        kmer_correction = -1*(plot_utils.search_for_profile_base_shift(args.profile)[0] + plot_utils.search_for_profile_base_shift(args.profile)[1])

    sig_algn_dic = {}

    y_shift = 0
    prev_y_max = 0
    prev_y_min = 0
    p = plot_utils.create_figure(args, plot_mode=1)
    previous_plot = p

    if use_paf == 1 and plot_sig_ref_flag == 0:
        raise Exception("Error: If you are trying to plot just reads please use plot.py. Otherwise provide the correct signal-reference alignment.")
    elif use_paf == 0 and plot_sig_ref_flag == 1: # using sam/bam
        print("Info: Signal to reference method using SAM/BAM ...")
        fasta_reads = Fasta(args.file)

        # check if there exists a .bam.bai
        index_file = args.alignment + ".bai"
        if not os.path.exists(index_file):
            raise Exception("Error: please provide a bam file that is sorted and indexed to extract the regions")

        args_region = re.sub(',', '', args.region)
        # print(args_region)
        # pattern = re.compile("^[a-z]+[0-9]+\:[0-9]+\-[0-9]+")
        pattern = re.compile("^.*\:[0-9]+\-[0-9]+")
        if not pattern.match(args_region):
            raise Exception("Error: region provided is not in correct format")
        args_ref_name = args_region.split(":")[0]
        args_ref_start = int(args_region.split(":")[1].split("-")[0])
        args_ref_end = int(args_region.split(":")[1].split("-")[1])

        samfile = pysam.AlignmentFile(args.alignment, mode='rb')
        # if args.bed:
        #     p = bed_annotation.plot_bed_annotation(p=p, ref_id=args_ref_name, bed_dic=bed_dic, sig_algn_data=sig_algn_dic, draw_data=draw_data, base_limit=base_limit)

        for sam_record in samfile.fetch(contig=args_ref_name, start=args_ref_start, stop=args_ref_end):
            if args_ref_name != sam_record.reference_name:
                raise Exception("Error: sam record's reference name [" + sam_record.reference_name + "] and the name specified are different [" + ref_name + "]")
            read_id = sam_record.query_name
            if sam_record.is_supplementary or sam_record.is_unmapped or sam_record.is_secondary:
                continue
            if args.plot_reverse is True and sam_record.is_reverse is False:
                continue
            if args.plot_reverse is False and sam_record.is_reverse is True:
                continue
            if args.read_id != "" and read_id != args.read_id:
                continue
            if args.read_list != "" and read_id not in read_id_list:
                continue

            ref_seq_len = 0
            data_is_rna = 0
            start_index = -1
            if sam_record.has_tag("si"):
                si_tag = sam_record.get_tag("si").split(',')
                start_index = int(si_tag[SI_START_RAW])
                end_index = int(si_tag[SI_END_RAW])
                ref_seq_len = int(si_tag[SI_END_KMER]) - int(si_tag[SI_START_KMER])
                reference_start = int(si_tag[SI_START_KMER])
                if int(si_tag[SI_START_KMER]) > int(si_tag[SI_END_KMER]):  # if RNA start_kmer>end_kmer in paf
                    data_is_rna = 1
                    if not args.rna:
                        print("Info: data is detected as RNA")
                        raise Exception("Error: data is not specified as RNA. Please provide the argument --rna ")
                    ref_seq_len = int(si_tag[SI_START_KMER]) - int(si_tag[SI_END_KMER])
                    reference_start = int(si_tag[SI_END_KMER]) + kmer_correction

            else:
                raise Exception("Error: sam record does not have a 'si' tag.")
            # print("ref_seq_len: " + str(ref_seq_len))
            if ref_seq_len < BASE_LIMIT:
                base_limit = ref_seq_len
            else:
                base_limit = BASE_LIMIT
            sam_record_reference_end = reference_start + ref_seq_len #1based closed
            if not args.loose_bound:
                if data_is_rna == 1:
                    if args_ref_start < reference_start + 1 - kmer_correction:
                        continue
                else:
                    if args_ref_start < reference_start + 1:
                        continue
                if args_ref_end > sam_record_reference_end:
                    continue

            ref_name = sam_record.reference_name
            ref_start = reference_start + 1
            ref_end = ref_start + base_limit - 1 #ref_end is 1based closed
            if args_ref_start > ref_start:
                ref_start = args_ref_start
                if (ref_start + base_limit - 1) < sam_record_reference_end:
                    ref_end = ref_start + base_limit - 1
                else:
                    ref_end = sam_record_reference_end
            if args_ref_end < ref_end:
                ref_end = args_ref_end
            if ref_end < ref_start:
                print("Warning: a corner case has hit because the kmer_length used is larger than 1. This alignment will be skipped")
                continue
            base_limit = ref_end - ref_start + 1

            # print("ref_start: {}".format(ref_start))
            # print("ref_end: {}".format(ref_end))
            # print("ref_seq_len: {}".format(ref_seq_len))
            # print("{}".format(sam_record.cigarstring))
            # print("base_limit: {}".format(base_limit))
            if sam_record.is_reverse and data_is_rna == 1:
                raise Exception("Error: A transcript is  always sequenced from 3` to 5`. Squigualiser only supports reads mapped to the transcriptome.")

            if data_is_rna == 1:
                print("plot (RNA 3'->5') region: {}:{}-{}\tread_id: {}".format(ref_name, ref_end, ref_start, read_id))
                fasta_seq = fasta_reads.get_seq(name=ref_name, start=ref_start, end=ref_end).seq
            else:
                if sam_record.is_reverse:
                    print("plot (-) region: {}:{}-{}\tread_id: {}".format(ref_name, ref_start, ref_end, read_id))
                    fasta_seq = fasta_reads.get_seq(name=ref_name, start=ref_start, end=ref_end).seq
                    nn = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'}
                    # fasta_seq = "".join(nn[n] for n in reversed(fasta_seq))
                    fasta_seq = "".join(nn[n] for n in fasta_seq)
                else:
                    print("plot (+) region: {}:{}-{}\tread_id: {}".format(ref_name, ref_start, ref_end, read_id))
                    fasta_seq = fasta_reads.get_seq(name=ref_name, start=ref_start, end=ref_end).seq

            x = []
            x_real = []
            y = []

            read = s5.get_read(read_id, pA=args.no_pa, aux=["read_number", "start_mux"])
            if read is not None:
                # print("read_id:", read['read_id'])
                # print("len_raw_signal:", read['len_raw_signal'])
                # end_index = read['len_raw_signal']
                x = list(range(1, end_index - start_index + 1))
                x_real = list(range(start_index+1, end_index+1))             # 1based
                y = read['signal'][start_index:end_index]

            y = plot_utils.scale_signal(y, args.sig_scale)
            scaling_str = "no scaling"
            if args.sig_scale == "medmad" or args.sig_scale == "znorm":
                scaling_str = args.sig_scale
                draw_data["plot_y_margin"] = 0.1
            elif not args.sig_scale == "":
                raise Exception("Error: given --sig_scale method: {} is not supported".format(args.sig_scale))

            moves_string = sam_record.get_tag("ss")
            moves_string = re.sub('D', 'D,', moves_string)
            moves_string = re.sub('I', 'I,', moves_string).rstrip(',')
            moves = re.split(r',+', moves_string)

            if data_is_rna == 0:
                strand_dir = "(DNA +)"
                if sam_record.is_reverse:
                    strand_dir = "(DNA -)"
                    x_real.reverse()
                    y = np.flip(y)
                    moves.reverse()
            if data_is_rna == 1:
                strand_dir = "(RNA 3'->5')"
                fasta_seq = fasta_seq[::-1]
                if sam_record.is_reverse:
                    raise Exception("Error: data is rna and sam record is reverse mapped. This is not implemented yet. Please report")

            signal_tuple = (x, x_real, y)
            region_tuple = (ref_start, ref_end, reference_start, reference_start+ref_seq_len)

            sig_algn_dic['start_kmer'] = 0
            sig_algn_dic['ref_start'] = ref_start
            sig_algn_dic['ref_end'] = ref_end
            sig_algn_dic['pa'] = args.no_pa
            sig_algn_dic['plot_sig_ref_flag'] = plot_sig_ref_flag
            sig_algn_dic['data_is_rna'] = data_is_rna

            sig_algn_dic['ss'] = moves
            # print(len(moves))
            # print(fasta_seq)
            signal_tuple, region_tuple, sig_algn_dic, fasta_seq = plot_utils.adjust_before_plotting(ref_seq_len, signal_tuple, region_tuple, sig_algn_dic, fasta_seq)
            # print(len(sig_algn_dic['ss']))
            # if args.auto_base_shift and num_plots == 0:
            #     draw_data["base_shift"] = plot_utils.calculate_base_shift(signal_tuple[2], fasta_seq, sig_algn_dic['ss'], args)
            #     print("automatically calculated base_shift: {}".format(draw_data["base_shift"]))

            if draw_data["base_shift"] < 0:
                abs_base_shift = abs(draw_data["base_shift"])
                x = signal_tuple[0]
                x_real = signal_tuple[1]
                y = signal_tuple[2]
                y_prefix = [np.nan] * abs_base_shift * draw_data["fixed_base_width"]
                y = np.concatenate((y_prefix, y), axis=0)
                x_real = np.concatenate(([1] * abs_base_shift * draw_data["fixed_base_width"], x_real), axis=0)
                x = list(range(1, len(x) + 1 + abs_base_shift * draw_data["fixed_base_width"]))
                signal_tuple = (x, x_real, y)
                moves_prefix = [str(draw_data["fixed_base_width"])] * abs_base_shift
                sig_algn_dic['ss'] = moves_prefix + sig_algn_dic['ss']

            sig_algn_dic['tag_name'] = args.tag_name + indt + "base_shift: " + str(draw_data["base_shift"]) + indt + "scale:" + scaling_str + indt + "fixed_width: " + str(args.base_width) + indt + strand_dir + indt + "region: " + ref_name + ":"

            y_min = math.floor(np.nanmin(y))
            y_max = math.ceil(np.nanmax(y))

            if num_plots == 0 and args.bed and args.overlap_bottom is False:
                draw_data['y_min'] = y_min
                draw_data['y_max'] = y_max
                draw_data['fixed_width'] = True
                p = bed_annotation.plot_bed_annotation(p=p, ref_id=args_ref_name, bed_dic=bed_dic, sig_algn_data=sig_algn_dic, draw_data=draw_data, base_limit=base_limit)

            if not args.no_overlap and not args.overlap_only:
                if num_plots == 0:
                    p, location_plot, base_index = plot_function_fixed_width_pileup(read_id=read_id, signal_tuple=signal_tuple, sig_algn_data=sig_algn_dic, fasta_sequence=fasta_seq, base_limit=base_limit, draw_data=draw_data, p=previous_plot, num_plots=-1, y_shift=y_shift, y_min=y_min, y_max=y_max)
                    previous_plot = p
                    prev_y_max = y_max
                    prev_y_min = y_min
                    if max_base_index < base_index:
                        max_base_index = base_index
                    if max_location_plot < location_plot:
                        max_location_plot = location_plot

            if args.overlap_bottom:
                # y_shift = y_shift + prev_y_min + prev_y_max - prev_y_min + draw_data["plot_y_margin"] - y_min
                y_shift = y_shift + prev_y_max + draw_data["plot_y_margin"] - y_min
            else:
                y_shift = y_shift + prev_y_min - draw_data["plot_y_margin"] - y_max
            if args.overlap_only:
                y_shift = 0
            p, location_plot, base_index = plot_function_fixed_width_pileup(read_id=read_id, signal_tuple=signal_tuple, sig_algn_data=sig_algn_dic, fasta_sequence=fasta_seq, base_limit=base_limit, draw_data=draw_data, p=previous_plot, num_plots=num_plots, y_shift=y_shift, y_min=y_min, y_max=y_max)
            previous_plot = p
            prev_y_max = y_max
            prev_y_min = y_min
            if max_base_index < base_index:
                max_base_index = base_index
            if max_location_plot < location_plot:
                max_location_plot = location_plot
            num_plots += 1
            if num_plots == args.plot_limit:
                break
        if args.bed and args.overlap_bottom is True:
            draw_data['y_min'] = prev_y_min + y_shift
            draw_data['y_max'] = prev_y_max + y_shift
            draw_data['fixed_width'] = True
            p = bed_annotation.plot_bed_annotation(p=p, ref_id=args_ref_name, bed_dic=bed_dic, sig_algn_data=sig_algn_dic, draw_data=draw_data, base_limit=base_limit)

    elif use_paf == 1 and plot_sig_ref_flag == 1:
        print("Info: Signal to reference method using PAF ...")
        fasta_reads = Fasta(args.file)
        tbxfile = pysam.TabixFile(args.alignment)

        args_region = re.sub(',', '', args.region)
        # print(args_region)
        # pattern = re.compile("^[a-z]+[0-9]+\:[0-9]+\-[0-9]+")
        pattern = re.compile("^.*\:[0-9]+\-[0-9]+")
        if not pattern.match(args_region):
            raise Exception("Error: region provided is not in correct format")
        args_ref_name = args_region.split(":")[0]
        args_ref_start = int(args_region.split(":")[1].split("-")[0])
        args_ref_end = int(args_region.split(":")[1].split("-")[1])
        for paf_record in tbxfile.fetch(args_ref_name, args_ref_start, args_ref_end, parser=pysam.asTuple()):
            # if paf_record[READ_ID] == paf_record[SEQUENCE_ID]:
            #     raise Exception("Error: this paf file is a signal to read mapping.")
            if args_ref_name != paf_record[SEQUENCE_ID]:
                raise Exception("Error: sam record's reference name [" + paf_record[SEQUENCE_ID] + "] and the name specified are different [" + ref_name + "]")
            read_id = paf_record[READ_ID]
            if args.read_id != "" and read_id != args.read_id:
                continue
            if args.read_list != "" and read_id not in read_id_list:
                continue
            if args.plot_reverse is True and paf_record[STRAND] == "+":
                continue
            if args.plot_reverse is False and paf_record[STRAND] == "-":
                continue

            data_is_rna = 0
            start_index = int(paf_record[START_RAW])
            end_index = int(paf_record[END_RAW])
            ref_seq_len = int(paf_record[END_KMER]) - int(paf_record[START_KMER])
            reference_start = int(paf_record[START_KMER])
            if int(paf_record[START_KMER]) > int(paf_record[END_KMER]):  # if RNA start_kmer>end_kmer in paf
                data_is_rna = 1
                if not args.rna:
                    print("Info: data is detected as RNA")
                    raise Exception("Error: data is not specified as RNA. Please provide the argument --rna ")
                ref_seq_len = int(paf_record[START_KMER]) - int(paf_record[END_KMER])
                reference_start = int(paf_record[END_KMER]) + kmer_correction
            # print("ref_seq_len: " + str(ref_seq_len))
            if ref_seq_len < BASE_LIMIT:
                base_limit = ref_seq_len
            else:
                base_limit = BASE_LIMIT
            paf_record_reference_end = reference_start + ref_seq_len #1based closed
            if not args.loose_bound:
                if data_is_rna == 1:
                    if args_ref_start < reference_start + 1 - kmer_correction:
                        continue
                else:
                    if args_ref_start < reference_start + 1:
                        continue
                if args_ref_end > paf_record_reference_end:
                    continue

            ref_name = paf_record[SEQUENCE_ID]
            ref_start = reference_start + 1
            ref_end = ref_start + base_limit - 1 #ref_end is 1based closed
            if args_ref_start > ref_start:
                ref_start = args_ref_start
                if (ref_start + base_limit - 1) < paf_record_reference_end:
                    ref_end = ref_start + base_limit - 1
                else:
                    ref_end = paf_record_reference_end
            if args_ref_end < ref_end:
                ref_end = args_ref_end
            if ref_end < ref_start:
                print("Warning: a corner case has hit because the kmer_length used is larger than 1. This alignment will be skipped")
                continue
            base_limit = ref_end - ref_start + 1

            # print("ref_start: {}".format(ref_start))
            # print("ref_end: {}".format(ref_end))
            # print("ref_seq_len: {}".format(ref_seq_len))
            # print("base_limit: {}".format(base_limit))

            record_is_reverse = 0
            if paf_record[STRAND] == "-":
                record_is_reverse = 1
            if record_is_reverse and data_is_rna == 1:
                raise Exception("Error: A transcript is  always sequenced from 3` to 5`. Squigualiser only supports reads mapped to the transcriptome.")
            if data_is_rna == 1:
                print("plot (RNA 3'->5') region: {}:{}-{}\tread_id: {}".format(ref_name, ref_end, ref_start, read_id))
                fasta_seq = fasta_reads.get_seq(name=ref_name, start=ref_start, end=ref_end).seq
            else:
                if record_is_reverse:
                    print("plot (-) region: {}:{}-{}\tread_id: {}".format(ref_name, ref_start, ref_end, read_id))
                    fasta_seq = fasta_reads.get_seq(name=ref_name, start=ref_start, end=ref_end).seq
                    nn = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'}
                    # fasta_seq = "".join(nn[n] for n in reversed(fasta_seq))
                    fasta_seq = "".join(nn[n] for n in fasta_seq)
                else:
                    print("plot (+) region: {}:{}-{}\tread_id: {}".format(ref_name, ref_start, ref_end, read_id))
                    fasta_seq = fasta_reads.get_seq(name=ref_name, start=ref_start, end=ref_end).seq

            x = []
            x_real = []
            y = []

            read = s5.get_read(read_id, pA=args.no_pa, aux=["read_number", "start_mux"])
            if read is not None:
                # print("read_id:", read['read_id'])
                # print("len_raw_signal:", read['len_raw_signal'])
                # end_index = read['len_raw_signal']
                x = list(range(1, end_index - start_index + 1))
                x_real = list(range(start_index + 1, end_index + 1))  # 1based
                y = read['signal'][start_index:end_index]

            y = plot_utils.scale_signal(y, args.sig_scale)
            scaling_str = "no scaling"
            if args.sig_scale == "medmad" or args.sig_scale == "znorm":
                scaling_str = args.sig_scale
                draw_data["plot_y_margin"] = 0.1
            elif not args.sig_scale == "":
                raise Exception("Error: given --sig_scale method: {} is not supported".format(args.sig_scale))

            for i in range(12, len(paf_record)):
                tag = paf_record[i][:2]
                if tag == "ss":
                    moves_string = paf_record[i][5:]
            if moves_string == "":
                raise Exception("Error: ss tag was not found")
            moves_string = re.sub('D', 'D,', moves_string)
            moves_string = re.sub('I', 'I,', moves_string).rstrip(',')
            moves = re.split(r',+', moves_string)

            if data_is_rna == 0:
                strand_dir = "(DNA +)"
                if record_is_reverse:
                    strand_dir = "(DNA -)"
                    x_real.reverse()
                    y = np.flip(y)
                    moves.reverse()
            if data_is_rna == 1:
                strand_dir = "(RNA 3'->5')"
                fasta_seq = fasta_seq[::-1]

            signal_tuple = (x, x_real, y)
            region_tuple = (ref_start, ref_end, reference_start, reference_start + ref_seq_len)

            sig_algn_dic['start_kmer'] = 0
            sig_algn_dic['ref_start'] = ref_start
            sig_algn_dic['ref_end'] = ref_end
            sig_algn_dic['pa'] = args.no_pa
            sig_algn_dic['plot_sig_ref_flag'] = plot_sig_ref_flag
            sig_algn_dic['data_is_rna'] = data_is_rna

            sig_algn_dic['ss'] = moves
            # print(len(moves))
            # print(fasta_seq)
            signal_tuple, region_tuple, sig_algn_dic, fasta_seq = plot_utils.adjust_before_plotting(ref_seq_len, signal_tuple, region_tuple, sig_algn_dic, fasta_seq)

            # if args.auto_base_shift and num_plots == 0:
            #     draw_data["base_shift"] = plot_utils.calculate_base_shift(signal_tuple[2], fasta_seq, sig_algn_dic['ss'], args)
            #     print("automatically calculated base_shift: {}".format(draw_data["base_shift"]))

            if draw_data["base_shift"] < 0:
                abs_base_shift = abs(draw_data["base_shift"])
                x = signal_tuple[0]
                x_real = signal_tuple[1]
                y = signal_tuple[2]
                y_prefix = [np.nan] * abs_base_shift * draw_data["fixed_base_width"]
                y = np.concatenate((y_prefix, y), axis=0)
                x_real = np.concatenate(([1] * abs_base_shift * draw_data["fixed_base_width"], x_real), axis=0)
                x = list(range(1, len(x) + 1 + abs_base_shift * draw_data["fixed_base_width"]))
                signal_tuple = (x, x_real, y)
                moves_prefix = [str(draw_data["fixed_base_width"])] * abs_base_shift
                sig_algn_dic['ss'] = moves_prefix + sig_algn_dic['ss']

            sig_algn_dic['tag_name'] = args.tag_name + indt + "base_shift: " + str(draw_data["base_shift"]) + indt + "scale:" + scaling_str + indt + "fixed_width: " + str(args.base_width) + indt + strand_dir + indt + "region: " + ref_name + ":"

            # print(len(sig_algn_dic['ss']))
            y_min = math.floor(np.nanmin(y))
            y_max = math.ceil(np.nanmax(y))

            if num_plots == 0 and args.bed and args.overlap_bottom is False:
                draw_data['y_min'] = y_min
                draw_data['y_max'] = y_max
                draw_data['fixed_width'] = True
                p = bed_annotation.plot_bed_annotation(p=p, ref_id=args_ref_name, bed_dic=bed_dic, sig_algn_data=sig_algn_dic, draw_data=draw_data, base_limit=base_limit)

            if not args.no_overlap and not args.overlap_only:
                if num_plots == 0:
                    p, location_plot, base_index = plot_function_fixed_width_pileup(read_id=read_id, signal_tuple=signal_tuple, sig_algn_data=sig_algn_dic, fasta_sequence=fasta_seq, base_limit=base_limit, draw_data=draw_data, p=previous_plot, num_plots=-1, y_shift=y_shift, y_min=y_min, y_max=y_max)
                    previous_plot = p
                    prev_y_max = y_max
                    prev_y_min = y_min
                    if max_base_index < base_index:
                        max_base_index = base_index
                    if max_location_plot < location_plot:
                        max_location_plot = location_plot

            if args.overlap_bottom:
                # y_shift = y_shift + prev_y_min + prev_y_max - prev_y_min + draw_data["plot_y_margin"] - y_min
                y_shift = y_shift + prev_y_max + draw_data["plot_y_margin"] - y_min
            else:
                y_shift = y_shift + prev_y_min - draw_data["plot_y_margin"] - y_max
            if args.overlap_only:
                y_shift = 0
            p, location_plot, base_index = plot_function_fixed_width_pileup(read_id=read_id, signal_tuple=signal_tuple, sig_algn_data=sig_algn_dic, fasta_sequence=fasta_seq, base_limit=base_limit, draw_data=draw_data, p=previous_plot, num_plots=num_plots, y_shift=y_shift, y_min=y_min, y_max=y_max)
            previous_plot = p
            prev_y_max = y_max
            prev_y_min = y_min
            if max_base_index < base_index:
                max_base_index = base_index
            if max_location_plot < location_plot:
                max_location_plot = location_plot

            num_plots += 1
            if num_plots == args.plot_limit:
                break
        if args.bed and args.overlap_bottom is True:
            draw_data['y_min'] = prev_y_min + y_shift
            draw_data['y_max'] = prev_y_max + y_shift
            draw_data['fixed_width'] = True
            p = bed_annotation.plot_bed_annotation(p=p, ref_id=args_ref_name, bed_dic=bed_dic, sig_algn_data=sig_algn_dic, draw_data=draw_data, base_limit=base_limit)

    else:
        raise Exception("Error: You should not have ended up here. Please check your arguments")

    print("Number of plots: {}".format(num_plots))
    if num_plots > 0:
        if sig_algn_dic["data_is_rna"] == 1:
            sig_dir = " ->"
            plot_title = f'{sig_algn_dic["tag_name"]}[{sig_algn_dic["ref_end"]:,}-{sig_algn_dic["ref_end"] - max_base_index + 1:,}]{indt}num reads:{num_plots}{indt}signal dir:{sig_dir}'
        else:
            if args.plot_reverse:
                sig_dir = " <-"
            else:
                sig_dir = " ->"
            plot_title = f'{sig_algn_dic["tag_name"]}[{sig_algn_dic["ref_start"]:,}-{sig_algn_dic["ref_start"] + max_base_index - 1:,}]{indt}num reads:{num_plots}{indt}signal dir:{sig_dir}'
        p.title = plot_title
        p.legend.click_policy = "hide"
        # p.legend.location = 'top_left'
        p.legend.label_text_font_size = '7pt'
        p.legend.background_fill_alpha = 0.5

        if max_location_plot > y_shift:
            if max_location_plot > PLOT_X_RANGE:
                new_x_range = Range1d(0, PLOT_X_RANGE, bounds=(-1*PLOT_X_PADDING, max_location_plot+PLOT_X_PADDING))
            else:
                new_x_range = Range1d(0, max_location_plot, bounds=(-1*PLOT_X_PADDING, max_location_plot+PLOT_X_PADDING))
            new_x_range.js_property_callbacks = p.x_range.js_property_callbacks
            p.x_range = new_x_range

        renderer = p.multi_line([[1, 1]], [[1, 1]], line_width=4, alpha=0.4, color='black')
        draw_tool = FreehandDrawTool(renderers=[renderer], num_objects=50)
        p.add_tools(draw_tool)
        if args.return_plot:
            return p, num_plots
        else:
            pileup_output_file_name = args.output_dir + "/" + "pileup_" + args.tag_name + ".html"
            output_file(pileup_output_file_name, title="pileup_" + args.tag_name)
            print(f'output file: {os.path.abspath(pileup_output_file_name)}')
            save(p)
    elif num_plots == 0 and args.return_plot:
        return None, num_plots

    s5.close()

    if args.cprofile:
        pr.disable()
        s = io.StringIO()
        sort_by = 'cumulative'
        ps = pstats.Stats(pr, stream=s).sort_stats(sort_by)
        ps.print_stats()
        with open("cprofile_run.log", 'w') as f:
            print(s.getvalue(), file=f)
def argparser():
    # parser = argparse.ArgumentParser()
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False
    )

    parser.add_argument('-f', '--file', required=False, type=str, default="", help="fasta/fa/fastq/fq/fq.gz sequence file")
    parser.add_argument('-a', '--alignment', required=False, type=str, default="", help="for read-signal alignment use PAF\nfor reference-signal alignment use SAM/BAM")
    parser.add_argument('-s', '--slow5', required=False, type=str, default="", help="slow5 file")
    parser.add_argument('--region', required=False, type=str, default="", help="[start-end] 1-based closed interval region to plot. For SAM/BAM eg: chr1:6811428-6811467 or chr1:6,811,428-6,811,467. For PAF eg:100-200.")
    parser.add_argument('--tag_name', required=False, type=str, default="", help="a tag name to easily identify the plot")
    parser.add_argument('-r', '--read_id', required=False, type=str, default="", help="plot the read with read_id")
    parser.add_argument('-l', '--read_list', required=False, type=str, default="", help="a file with read_ids to plot")
    parser.add_argument('--base_limit', required=False, type=int, help="maximum number of bases to plot")
    parser.add_argument('--plot_reverse', required=False, action='store_true', help="plot only reverse mapped reads")
    parser.add_argument('--plot_num_samples', required=False, action='store_true', help="annotate the number of samples for each move")
    parser.add_argument('--rna', required=False, action='store_true', help="specify for RNA reads")
    parser.add_argument('--sig_scale', required=False, type=str, default="", help="plot the scaled signal. Supported scalings: [medmad, znorm]")
    parser.add_argument('--no_pa', required=False, action='store_false', help="skip converting the signal to pA values")
    parser.add_argument('--overlap_bottom', required=False, action='store_true', help="plot the overlap at the bottom")
    parser.add_argument('--no_overlap', required=False, action='store_true', help="skip plotting the overlap")
    parser.add_argument('--overlap_only', required=False, action='store_true', help="plot only the overlap")
    parser.add_argument('--loose_bound', required=False, action='store_true', help="also plot alignments not completely within the specified region")
    parser.add_argument('--point_size', required=False, type=int, default=0.5, help="signal point radius [0.5]")
    parser.add_argument('--base_width', required=False, type=int, default=FIXED_BASE_WIDTH, help="base width when plotting with fixed base width")
    parser.add_argument('--base_shift', required=False, type=int, default=PLOT_BASE_SHIFT, help="the number of bases to shift to align fist signal move")
    parser.add_argument('--profile', required=False, default="", type=str, help="determine base_shift using preset values")
    parser.add_argument('--list_profile', action='store_true', help="list the available profiles")
    parser.add_argument('--plot_limit', required=False, type=int, default=1000, help="limit the number of plots generated")
    parser.add_argument('--sig_plot_limit', required=False, type=int, default=SIG_PLOT_LENGTH, help="maximum number of signal samples to plot")
    parser.add_argument('--bed', required=False, help="bed file with annotations")
    parser.add_argument('--print_bed_labels',  required=False, action='store_true', help="draw bed annotations with labels")
    parser.add_argument('--cprofile', required=False, action='store_true', help="run cProfile for benchmarking")
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('-o', '--output_dir', type=str, default="", help="output dir")
    group.add_argument('--return_plot', action='store_true', help="return plot object without saving to output")
    return parser

if __name__ == "__main__":
    parser = argparser()
    args = parser.parse_args()
    try:
        run(args)
    except Exception as e:
        print(str(e))
        exit(1)




"""
Signal to seQuence alignment Plot - plot
Hiruna Samarakoon - Garvan Medical Institute
hiruna@unsw.edu.au
"""
import numpy as np
from bokeh.plotting import figure, show, output_file, save
from bokeh.models import Span, BoxAnnotation, HoverTool, WheelZoomTool, ColumnDataSource, Label, LabelSet, CustomJS
from bokeh.colors import RGB
import pyslow5
import copy
import argparse
import re
import logging

BASE_SHIFT = 5
BASE_LIMIT = 1000
FRONT_CLIP = 50
SIG_PLOT_LENGTH = 8000
TRIM_OFFSET = 0
base_color_map = {'A': 'limegreen', 'C': 'blue', 'T': 'red', 'G': 'orange'}

# # importing the modules
# from bokeh.plotting import figure, output_file, show
# # file to save the model
# output_file("gfg.html")
# # instantiating the figure object
# graph = figure(title="Bokeh Multi Line Graph")
# # the points to be plotted
# xs = [[1, 2, 3, 4, 5], [-4, -2, 0, 2, 4]]
# ys = [[5, 3, 8, 0], [5, -4, 10, -2, 5]]
# # plotting the graph
# graph.multi_line(xs, ys)
# # displaying the model
# show(graph)
# exit(0)

parser = argparse.ArgumentParser()
   
parser.add_argument('-f', '--fastq', required=False, help="fastq file")
parser.add_argument('-r', '--read_id', required=False, help="read id")
parser.add_argument('--trim_offset', required=False, help="signal trim offset")
parser.add_argument('--plot_signal_limit', required=False, help="signal plot length")
parser.add_argument('-s', '--slow5', required=True, help="slow5 file")
parser.add_argument('--second_slow5', required=True, help="second slow5 file")
parser.add_argument('--second_fastq', required=False, help="second fastq file")
parser.add_argument('-a', '--alignment', required=False, help="alignment file")
parser.add_argument('-b', '--base_shift', required=False, default=BASE_SHIFT, help="base shift")
parser.add_argument('-o', '--output', required=True, help="output file (html)")

args = parser.parse_args()

flag_read_id = 0
if args.read_id:
    flag_read_id = 1
    print(f'read_id: {args.read_id}')
    read_id = args.read_id

trim_offset = TRIM_OFFSET
if args.trim_offset:
    trim_offset = args.trim_offset

plot_signal_limit = SIG_PLOT_LENGTH
if args.plot_signal_limit:
    plot_signal_limit = args.plot_signal_limit

flag_fastq = 0
if args.fastq:
    flag_fastq = 1
    print(f'fastq file: {args.fastq}')
print(f'signal file: {args.slow5}')
flag_alignment = 0
if args.alignment:
    flag_alignment = 1
    print(f'alignment file: {args.alignment}')
print(f'base_shift: {args.base_shift}')
print(f'output file: {args.output}')

if flag_alignment == 1 and flag_fastq == 0:
    logging.error("please provide both fastq and alignment files")
    exit()

if flag_fastq == 1:
    # read fastq file
    fastq_file = open(args.fastq, 'r')
    read_id = fastq_file.readline().split()[0][1:]
    fastq_seq = fastq_file.readline().rstrip()
    if BASE_LIMIT > len(fastq_seq):
        BASE_LIMIT = len(fastq_seq)
    fastq_file.close()

if flag_alignment == 1:
    # read alignment file
    align_file = open(args.alignment, 'r')
    alignment = align_file.readline().split()
    trim_offset = int(alignment[2])
    moves_string = alignment[14]
    align_file.close()

# open signal file
x = []
x_real = []
y = [0]

s5 = pyslow5.Open(args.slow5, 'r')
read = s5.get_read(read_id, pA=True, aux=["read_number", "start_mux"])
if read is not None:
    print("read_id:", read['read_id'])
    print("len_raw_signal:", read['len_raw_signal'])
    # x = list(range(1,read['len_raw_signal']+1))
    # y = read['signal']

    chunk_offset = 0
    if trim_offset == 0:
        FRONT_CLIP = 0
    start_index = trim_offset + chunk_offset - FRONT_CLIP
    if read['len_raw_signal'] < start_index + SIG_PLOT_LENGTH:
        SIG_PLOT_LENGTH = read['len_raw_signal'] - start_index
        end_index = read['len_raw_signal']
    else:
        end_index = start_index + SIG_PLOT_LENGTH
    shift = end_index

    x = list(range(0, end_index-start_index + 1))
    x_real = list(range(start_index, end_index + 1))
    y.extend(read['signal'][start_index:end_index])
s5.close()
# set output to static HTML file
output_file(filename=args.output, title=read_id)

plot_title = f'{read_id}:{start_index}-{end_index}-{trim_offset}-{chunk_offset}'
tools_to_show = 'hover,box_zoom,pan,save,wheel_zoom'
p = figure(title=plot_title,
		   x_axis_label='signal index',
		   y_axis_label='signal value',
		   sizing_mode="stretch_width",
		   height=300,
		   output_backend="webgl",
		   x_range=(0, 750),
		   tools=tools_to_show)
    # tooltips=tool_tips)

p.toolbar.active_scroll = p.select_one(WheelZoomTool)

if flag_alignment == 1:
    base_x = []
    base_y = []
    base_label = []
    base_count = 0
    location_plot = trim_offset+chunk_offset - start_index
    previous_location = location_plot
    # draw moves
    moves_string = re.sub('ss:Z:', '', moves_string)
    moves_string = re.sub('D', 'D,', moves_string)
    moves_string = re.sub('I', 'I,', moves_string).rstrip(',')
    # print(moves_string)
    moves = re.split(r',+', moves_string)

    vlines = []
    BASE_SHIFT = int(args.base_shift)
    base_count = BASE_SHIFT

    for i in moves:
        previous_location = location_plot
        n_samples = 0
        if 'D' in i:
            i = re.sub('D', '', i)
            n_samples = int(i)
            prev_loc = previous_location
            for j in range(0, n_samples):
                base = fastq_seq[base_count]
                base_box = BoxAnnotation(left=prev_loc, right=prev_loc+5, fill_alpha=0.2, fill_color='white')
                p.add_layout(base_box)

                base_x.append(prev_loc)
                base_y.append(115)
                label = str(base) + "\t" + str(base_count + 1)
                base_label.append(label)

                prev_loc = prev_loc + 5
                x_end = x[-1]
                x = x + list(range(x_end, x_end+5))
                base_count = base_count + 1
            location_plot = prev_loc
            z = np.concatenate((y[:previous_location], [0] * n_samples * 5), axis=0)
            y = np.concatenate((z, y[previous_location:]), axis=0)
            # y = y[:previous_location] + y[previous_location:]

        elif 'I' in i:
            i = re.sub('I', '', i)
            n_samples = int(i)
            location_plot = location_plot + n_samples

            vline = Span(location=location_plot, dimension='height', line_color='red', line_width=1)
            vlines.append(vline)

        else:
            n_samples = int(i)
            location_plot = location_plot + n_samples

            base = fastq_seq[base_count]
            base_box = BoxAnnotation(left=previous_location, right=location_plot, fill_alpha=0.2, fill_color=base_color_map[base])
            p.add_layout(base_box)

            vline = Span(location=location_plot, dimension='height', line_color='red', line_width=1)
            vlines.append(vline)

            base_x.append(previous_location)
            base_y.append(115)
            label = str(base) + "\t" + str(base_count + 1)
            base_label.append(label)
            base_count = base_count + 1

        if base_count == BASE_LIMIT:
            break
        if location_plot > SIG_PLOT_LENGTH:
            break

    p.renderers.extend(vlines)

    base_annotation = ColumnDataSource(data=dict(base_x=base_x,
                                                 base_y=base_y,
                                                 base_label=base_label))

    base_annotation_labels = LabelSet(x='base_x', y='base_y', text='base_label',
                                      x_offset=5, y_offset=5, source=base_annotation, render_mode='canvas',
                                      text_font_size="7pt")

    p.add_layout(base_annotation_labels)

    plot_signal_limit = location_plot + 10

y_second = [0]


if flag_fastq == 1:
    # read fastq file
    fastq_file = open(args.second_fastq, 'r')
    read_id = fastq_file.readline().split()[0][1:]
    fastq_seq = fastq_file.readline().rstrip()
    if BASE_LIMIT > len(fastq_seq):
        BASE_LIMIT = len(fastq_seq)
    fastq_file.close()

s5 = pyslow5.Open(args.second_slow5, 'r')
read = s5.get_read(read_id, pA=True)
if read is not None:
    print("read_id:", read['read_id'])
    print("len_raw_signal:", read['len_raw_signal'])
    # x = list(range(1,read['len_raw_signal']+1))
    # y = read['signal']
    chunk_offset = 0
    if trim_offset == 0:
        FRONT_CLIP = 0
    start_index = trim_offset + chunk_offset - FRONT_CLIP
    if read['len_raw_signal'] < start_index + SIG_PLOT_LENGTH:
        SIG_PLOT_LENGTH = read['len_raw_signal'] - start_index
        end_index = read['len_raw_signal']
    else:
        end_index = start_index + SIG_PLOT_LENGTH
    shift = end_index

    y_second.extend(read['signal'][start_index:end_index])
s5.close()

SIGNAL_SHIFT = 15
y_second = [value + SIGNAL_SHIFT for value in y_second]

p.line(x, y, line_width=2, color='black', legend_label='nanopolsih')
p.line(x, y_second, line_width=1, color='grey', legend_label='a14')
# add a circle renderer with a size, color, and alpha
# p.circle(x[:plot_signal_limit], y[:plot_signal_limit], size=2, color="red", alpha=0.5)

# show the tooltip
hover = p.select(dict(type=HoverTool))
hover.tooltips = [("x", "@x_real"), ("y", "$y")]
hover.mode = 'mouse'

save(p)



import numpy as np
from bokeh.io import output_file, show
from bokeh.layouts import gridplot
from bokeh.plotting import figure, save
from bokeh.models import Span, BoxAnnotation, HoverTool, WheelZoomTool, ColumnDataSource, Label, LabelSet, CustomJS
import argparse

from numpy import array, linspace
from sklearn.neighbors import KernelDensity
from scipy.signal import argrelextrema

FILLING_SIZE = 30

def main():

    flag_fastq_on = 0

    parser = argparse.ArgumentParser()

    parser.add_argument('-k', '--kmer', required=True, help="kmer")
    parser.add_argument('-d', '--kmer_dump', required=True, help="kmer dump file")
    parser.add_argument('-m', '--signal_print_margin', required=True, help="signal print margin used in poregen")
    parser.add_argument('-o', '--output', required=True, help="output file (html)")
    parser.add_argument('-r', '--fastq', required=False, help="fastq file (optional)")
    parser.add_argument('-n', '--ngb', required=False, help="no. of bases to the right (optional)")

    args = parser.parse_args()
    print(f'kmer: {args.kmer}')
    print(f'kmer_dump: {args.kmer_dump}')
    print(f'signal_print_margin: {args.signal_print_margin}')
    print(f'output file: {args.output}')

    kmer_size = len(args.kmer)
    kmer_neigbours_right = []

    num_neighbours = 1
    if args.ngb:
        num_neighbours = int(args.ngb)

    if args.fastq:
        print(f'fastq file: {args.fastq}')
        flag_fastq_on = 1
        fastq_file = open(args.fastq, 'r')
        fastq_file.readline()
        fastq_seq = fastq_file.readline().rstrip()
        shift_count = 0
        fastq_len = len(fastq_seq)
        for s in fastq_seq[kmer_size:]:
            kmer_ = fastq_seq[shift_count:shift_count+kmer_size]
            if kmer_ == args.kmer and shift_count+kmer_size+num_neighbours < fastq_len:
                bases_to_right = fastq_seq[shift_count+kmer_size+1:shift_count+kmer_size+1+num_neighbours]
                kmer_neigbours_right.append(bases_to_right)
            shift_count = shift_count + 1


    output_file(filename=args.output, title=args.kmer)

    # read kmer_dump file
    kmer_dump_file = open(args.kmer_dump, 'r')
    kmer_dump = kmer_dump_file.readline().replace(':', '').split(';')[:-1]
    kmer_dump_file.close()

    n_kmer_dump = len(kmer_dump)
    print(f'number of plots : {n_kmer_dump}')

    filling = [0] * FILLING_SIZE
    signal_print_margin = int(args.signal_print_margin)

    kmer_size=len(args.kmer)

    plot_title = f'{kmer_size}-mer:{args.kmer}\t\t\tno. of plots:{n_kmer_dump}\t\t\tgap between siglets:{FILLING_SIZE}'
    tools_to_show = 'hover,box_zoom,pan,save,wheel_zoom'
    p = figure(title=plot_title,
               y_axis_label='signal value',
               sizing_mode="stretch_width",
               height=300,
               output_backend="webgl",
               x_range=(0, 750),
               tools=tools_to_show)
        # tooltips=tool_tips)

    p.toolbar.active_scroll = p.select_one(WheelZoomTool)

    x = [] # x values: margin, annotation, gap
    y = [] # y values: margin, annoration, gap

    neighbour_x = []
    neighbour_y = []

    start_count = 0

    guppy_y_all = [] # list to contain all annotated valued

    for i in kmer_dump:

        # y_sig = i.split(',')
        y_sig = [float(i) for i in i.split(',')]
        len_i = len(y_sig)
        x_sig = list(range(start_count,  start_count + len_i + FILLING_SIZE))
        start_count = start_count + len_i + FILLING_SIZE

        x = x + x_sig
        y = y + y_sig + filling

        guppy_x = x_sig[signal_print_margin:-(signal_print_margin+FILLING_SIZE)]
        guppy_y = y_sig[signal_print_margin:-signal_print_margin]
        guppy_y_all = guppy_y_all + guppy_y
        p.circle(guppy_x, guppy_y, size=5, color="red", alpha=0.5)
        calculate_current_mean = method_0(guppy_y)
        current_line = [calculate_current_mean for j in guppy_x]
        p.line(guppy_x, current_line, color='black', line_width=2, line_dash='dotdash')

        if flag_fastq_on:
            x_value = guppy_x[-1]
            neighbour_x.append(x_value)
            neighbour_y.append(120)

    # print(guppy_y_all)
    cluster_median = method_1(guppy_y_all)
    for median in cluster_median:
        p.line(x, median, color='grey', line_width=2, line_dash='dotdash')

    source_1 = ColumnDataSource(data=dict(x=neighbour_x,
                                        y=neighbour_y,
                                        bases=kmer_neigbours_right))
    labels = LabelSet(x='x', y='y', text='bases',
                      x_offset=5, y_offset=5, source=source_1)
    p.add_layout(labels)
    # print(x)
    # print(y)
    # print(len(x))
    # print(len(y))
    # p.circle(x, y, size=5, color="red", alpha=0.5)

    source = ColumnDataSource(data=dict(
        x=x,
        y=y,
    ))
    p.line('x', 'y', line_width=2, source=source)

    # show the tooltip
    hover = p.select(dict(type=HoverTool))
    hover.tooltips = [("x", "$x"), ("y", "$y")]
    hover.mode = 'mouse'

    save(p)


def method_0(y):
    med = np.median(y)
    return med

def method_1(y):
    cluster_median = []
    a = np.array(y).reshape(-1, 1)
    # a = samples.reshape(-1,1)
    kde = KernelDensity(kernel='gaussian', bandwidth=3).fit(a)
    s = linspace(0, 100)
    e = kde.score_samples(s.reshape(-1, 1))
    mi, ma = argrelextrema(e, np.less)[0], argrelextrema(e, np.greater)[0]

    if len(mi) == 0:
        print("one cluster")
        cluster_median.append(np.median(y))
    elif len(mi) == 1:
        print("two clusters")
        cluster = a[a < s[mi][0]]
        cluster_median.append(np.median(cluster))
        cluster = a[a >= s[mi][-1]]
        cluster_median.append(np.median(cluster))
    else:
        print("more than two clusters")
        cluster_median.append(np.median(y))

    return cluster_median

main()


import setuptools
from os import path
import re

PKG_NAME = "squigualiser"
MOD_NAME = "src"

# add readme to long description as that's what pypi sees
with open("README.md", "r") as f:
    long_description = f.read()

# get version from file rather than here so change isn't in this file
__version__ = ""
exec(open("{}/_version.py".format(MOD_NAME)).read())

# User can set version of ont-pyguppy-client-lib to match guppy version
with open(path.join(path.abspath(path.dirname(__file__)), "requirements.txt")) as f:
    install_requires = [p.strip() for p in f]


setuptools.setup(
    name=PKG_NAME,
    version=__version__,
    url="https://github.com/hiruna72/squigualiser",
    author="Hiruna Samarakoon",
    author_email="h.samarakoon@garvan.org.au",
    maintainer='Hiruna Samarakoon',
    maintainer_email='h.samarakoon@garvan.org.au',
    description="Visualise ONT raw signals ",
    license='MIT',
    keywords=['nanopore', 'slow5', 'signal'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    python_requires=">=3.8",
    install_requires=install_requires,
    setup_requires=["numpy"],
    entry_points={
        "console_scripts": [
            "squigualiser={}.__init__:main".format(MOD_NAME),
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
)
